# -*- coding: utf-8 -*-
"""
markupsafe
~~~~~~~~~~

Implements an escape function and a Markup string to replace HTML
special characters with safe representations.

:copyright: 2010 Pallets
:license: BSD-3-Clause
"""
import re
import string

from ._compat import int_types
from ._compat import iteritems
from ._compat import Mapping
from ._compat import PY2
from ._compat import string_types
from ._compat import text_type
from ._compat import unichr

__version__ = "1.1.1"

__all__ = ["Markup", "soft_unicode", "escape", "escape_silent"]

_striptags_re = re.compile(r"(<!--.*?-->|<[^>]*>)")
_entity_re = re.compile(r"&([^& ;]+);")


class Markup(text_type):
    """A string that is ready to be safely inserted into an HTML or XML
    document, either because it was escaped or because it was marked
    safe.

    Passing an object to the constructor converts it to text and wraps
    it to mark it safe without escaping. To escape the text, use the
    :meth:`escape` class method instead.

    >>> Markup('Hello, <em>World</em>!')
    Markup('Hello, <em>World</em>!')
    >>> Markup(42)
    Markup('42')
    >>> Markup.escape('Hello, <em>World</em>!')
    Markup('Hello &lt;em&gt;World&lt;/em&gt;!')

    This implements the ``__html__()`` interface that some frameworks
    use. Passing an object that implements ``__html__()`` will wrap the
    output of that method, marking it safe.

    >>> class Foo:
    ...     def __html__(self):
    ...         return '<a href="/foo">foo</a>'
    ...
    >>> Markup(Foo())
    Markup('<a href="/foo">foo</a>')

    This is a subclass of the text type (``str`` in Python 3,
    ``unicode`` in Python 2). It has the same methods as that type, but
    all methods escape their arguments and return a ``Markup`` instance.

    >>> Markup('<em>%s</em>') % 'foo & bar'
    Markup('<em>foo &amp; bar</em>')
    >>> Markup('<em>Hello</em> ') + '<foo>'
    Markup('<em>Hello</em> &lt;foo&gt;')
    """

    __slots__ = ()

    def __new__(cls, base=u"", encoding=None, errors="strict"):
        if hasattr(base, "__html__"):
            base = base.__html__()
        if encoding is None:
            return text_type.__new__(cls, base)
        return text_type.__new__(cls, base, encoding, errors)

    def __html__(self):
        return self

    def __add__(self, other):
        if isinstance(other, string_types) or hasattr(other, "__html__"):
            return self.__class__(super(Markup, self).__add__(self.escape(other)))
        return NotImplemented

    def __radd__(self, other):
        if hasattr(other, "__html__") or isinstance(other, string_types):
            return self.escape(other).__add__(self)
        return NotImplemented

    def __mul__(self, num):
        if isinstance(num, int_types):
            return self.__class__(text_type.__mul__(self, num))
        return NotImplemented

    __rmul__ = __mul__

    def __mod__(self, arg):
        if isinstance(arg, tuple):
            arg = tuple(_MarkupEscapeHelper(x, self.escape) for x in arg)
        else:
            arg = _MarkupEscapeHelper(arg, self.escape)
        return self.__class__(text_type.__mod__(self, arg))

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, text_type.__repr__(self))

    def join(self, seq):
        return self.__class__(text_type.join(self, map(self.escape, seq)))

    join.__doc__ = text_type.join.__doc__

    def split(self, *args, **kwargs):
        return list(map(self.__class__, text_type.split(self, *args, **kwargs)))

    split.__doc__ = text_type.split.__doc__

    def rsplit(self, *args, **kwargs):
        return list(map(self.__class__, text_type.rsplit(self, *args, **kwargs)))

    rsplit.__doc__ = text_type.rsplit.__doc__

    def splitlines(self, *args, **kwargs):
        return list(map(self.__class__, text_type.splitlines(self, *args, **kwargs)))

    splitlines.__doc__ = text_type.splitlines.__doc__

    def unescape(self):
        """Convert escaped markup back into a text string. This replaces
        HTML entities with the characters they represent.

        >>> Markup('Main &raquo; <em>About</em>').unescape()
        'Main » <em>About</em>'
        """
        from ._constants import HTML_ENTITIES

        def handle_match(m):
            name = m.group(1)
            if name in HTML_ENTITIES:
                return unichr(HTML_ENTITIES[name])
            try:
                if name[:2] in ("#x", "#X"):
                    return unichr(int(name[2:], 16))
                elif name.startswith("#"):
                    return unichr(int(name[1:]))
            except ValueError:
                pass
            # Don't modify unexpected input.
            return m.group()

        return _entity_re.sub(handle_match, text_type(self))

    def striptags(self):
        """:meth:`unescape` the markup, remove tags, and normalize
        whitespace to single spaces.

        >>> Markup('Main &raquo;\t<em>About</em>').striptags()
        'Main » About'
        """
        stripped = u" ".join(_striptags_re.sub("", self).split())
        return Markup(stripped).unescape()

    @classmethod
    def escape(cls, s):
        """Escape a string. Calls :func:`escape` and ensures that for
        subclasses the correct type is returned.
        """
        rv = escape(s)
        if rv.__class__ is not cls:
            return cls(rv)
        return rv

    def make_simple_escaping_wrapper(name):  # noqa: B902
        orig = getattr(text_type, name)

        def func(self, *args, **kwargs):
            args = _escape_argspec(list(args), enumerate(args), self.escape)
            _escape_argspec(kwargs, iteritems(kwargs), self.escape)
            return self.__class__(orig(self, *args, **kwargs))

        func.__name__ = orig.__name__
        func.__doc__ = orig.__doc__
        return func

    for method in (
        "__getitem__",
        "capitalize",
        "title",
        "lower",
        "upper",
        "replace",
        "ljust",
        "rjust",
        "lstrip",
        "rstrip",
        "center",
        "strip",
        "translate",
        "expandtabs",
        "swapcase",
        "zfill",
    ):
        locals()[method] = make_simple_escaping_wrapper(method)

    def partition(self, sep):
        return tuple(map(self.__class__, text_type.partition(self, self.escape(sep))))

    def rpartition(self, sep):
        return tuple(map(self.__class__, text_type.rpartition(self, self.escape(sep))))

    def format(self, *args, **kwargs):
        formatter = EscapeFormatter(self.escape)
        kwargs = _MagicFormatMapping(args, kwargs)
        return self.__class__(formatter.vformat(self, args, kwargs))

    def __html_format__(self, format_spec):
        if format_spec:
            raise ValueError("Unsupported format specification " "for Markup.")
        return self

    # not in python 3
    if hasattr(text_type, "__getslice__"):
        __getslice__ = make_simple_escaping_wrapper("__getslice__")

    del method, make_simple_escaping_wrapper


class _MagicFormatMapping(Mapping):
    """This class implements a dummy wrapper to fix a bug in the Python
    standard library for string formatting.

    See http://bugs.python.org/issue13598 for information about why
    this is necessary.
    """

    def __init__(self, args, kwargs):
        self._args = args
        self._kwargs = kwargs
        self._last_index = 0

    def __getitem__(self, key):
        if key == "":
            idx = self._last_index
            self._last_index += 1
            try:
                return self._args[idx]
            except LookupError:
                pass
            key = str(idx)
        return self._kwargs[key]

    def __iter__(self):
        return iter(self._kwargs)

    def __len__(self):
        return len(self._kwargs)


if hasattr(text_type, "format"):

    class EscapeFormatter(string.Formatter):
        def __init__(self, escape):
            self.escape = escape

        def format_field(self, value, format_spec):
            if hasattr(value, "__html_format__"):
                rv = value.__html_format__(format_spec)
            elif hasattr(value, "__html__"):
                if format_spec:
                    raise ValueError(
                        "Format specifier {0} given, but {1} does not"
                        " define __html_format__. A class that defines"
                        " __html__ must define __html_format__ to work"
                        " with format specifiers.".format(format_spec, type(value))
                    )
                rv = value.__html__()
            else:
                # We need to make sure the format spec is unicode here as
                # otherwise the wrong callback methods are invoked.  For
                # instance a byte string there would invoke __str__ and
                # not __unicode__.
                rv = string.Formatter.format_field(self, value, text_type(format_spec))
            return text_type(self.escape(rv))


def _escape_argspec(obj, iterable, escape):
    """Helper for various string-wrapped functions."""
    for key, value in iterable:
        if hasattr(value, "__html__") or isinstance(value, string_types):
            obj[key] = escape(value)
    return obj


class _MarkupEscapeHelper(object):
    """Helper for Markup.__mod__"""

    def __init__(self, obj, escape):
        self.obj = obj
        self.escape = escape

    def __getitem__(self, item):
        return _MarkupEscapeHelper(self.obj[item], self.escape)

    def __str__(self):
        return text_type(self.escape(self.obj))

    __unicode__ = __str__

    def __repr__(self):
        return str(self.escape(repr(self.obj)))

    def __int__(self):
        return int(self.obj)

    def __float__(self):
        return float(self.obj)


# we have to import it down here as the speedups and native
# modules imports the markup type which is define above.
try:
    from ._speedups import escape, escape_silent, soft_unicode
except ImportError:
    from ._native import escape, escape_silent, soft_unicode

if not PY2:
    soft_str = soft_unicode
    __all__.append("soft_str")



import json
from typing import Union

import persistence
import pyscripts
import sysactions

mb_context = pyscripts.mbcontext
_modifiers = None
_sub_modifiers = {}

MODIFIERS_QUERY = """
SELECT DISTINCT *
FROM (
    SELECT 
        P1.ProductCode AS productCode,
        TRIM(REPLACE(P1.ProductName, '"', '')) AS productName,
        P2.ProductCode AS modifierCode,
        TRIM(REPLACE(P2.ProductName, '"', '')) AS modifierName,
        COALESCE((
                SELECT 1
                FROM ProductClassification
                WHERE ClassCode = P2.ProductCode LIMIT 1
                ), 0) AS isOption,
        COALESCE((
                SELECT 1
                FROM ProductKernelParams
                WHERE ProductCode = P2.ProductCode
                    AND ProductType = 2 LIMIT 1
                ), 0) AS isCombo,
        COALESCE((
                SELECT 1
                FROM ProductCustomParams
                WHERE ProductCode = P2.ProductCode
                    AND CustomParams.CustomParamId = 'CFH' COLLATE NOCASE
                    AND CustomParams.CustomParamValue = 'true' COLLATE NOCASE LIMIT 1
                ), 0) AS isCFH,
        (
            SELECT ('[' || COALESCE(GROUP_CONCAT('[' ||
                PC.ProductCode || ',"' ||
                TRIM(REPLACE(PPC.ProductName, '"', '')) || '", ' ||
                TRIM(
                    CASE WHEN PC.CustomAttr IS NULL OR PC.CustomAttr = ''
                    THEN '{}'
                    ELSE PC.CustomAttr
                    END
                ) ||']') , '') ||
                ']')
            FROM ProductClassification PC
            INNER JOIN Product PPC ON PPC.ProductCode = PC.ProductCode
            WHERE PC.ClassCode = P2.ProductCode
            ) AS options,
        PP.MinQty AS minQty,
        PP.MaxQty AS maxQty,
        PP.DefaultQty AS defaultQty,
        PP.IncludedQty AS includedQty,
        PP.CustomAttr AS customAttr
    FROM ProductPart PP
    INNER JOIN Product P1 ON P1.ProductCode = PP.ProductCode
    INNER JOIN Product P2 ON P2.ProductCode = PP.PartCode
    LEFT JOIN ProductCustomParams CustomParams ON CustomParams.ProductCode = PP.PartCode
    ) T
ORDER BY productCode, modifierCode
"""


@sysactions.action
def get_modifiers_data(_):
    return json.dumps(get_cached_modifiers())


def get_cached_modifiers():
    global _modifiers
    global _sub_modifiers

    if _modifiers is None:
        _sub_modifiers = {}
        conn = persistence.Driver().open(mb_context)
        modifiers = {}

        cursor = conn.select(MODIFIERS_QUERY)
        if cursor.rows() > 0:
            for row in cursor:
                get_modifier_product_data(_sub_modifiers, cursor, modifiers, row)

        conn.close()
        _modifiers = modifiers

    return {'modifiers': _modifiers, 'descriptions': _sub_modifiers}


# noinspection PyBroadException
def get_modifier_product_data(_sub_modifiers, cursor, modifiers, row):
    prod_code, prod_name, mod_code, mod_name, is_option, is_combo, is_cfh, options = map(row.get_entry, range(8))
    part_data = dict(zip(map(cursor.get_name, range(8, 13)), map(row.get_entry, range(8, 13))))

    if prod_code not in modifiers:
        modifiers[prod_code] = {'name': prod_name, 'parts': {}}

    if mod_code not in _sub_modifiers:
        _sub_modifiers[str(mod_code)] = mod_name

    if int(is_option) > 0 and options is not None:
        part_data['options'] = []
        options = options.replace("\\", "\\\\")
        for p_code, p_name, p_custom_attributes in json.loads(options):
            hide_on_pos_type = p_custom_attributes["hideOnPosType"] if "hideOnPosType" in p_custom_attributes else []
            part_sequencer = p_custom_attributes["partSequencer"] if "partSequencer" in p_custom_attributes else []
            part_data['options'].append(
                {"code": int(p_code), "hideOnPosType": hide_on_pos_type, "partSequencer": part_sequencer}
            )
            _sub_modifiers[str(p_code)] = p_name

    if 'customAttr' in part_data:
        try:
            if part_data['customAttr'] is not None:
                part_data['customAttr'] = json.loads(part_data['customAttr'])
        except Exception as _:
            pass
    else:
        part_data['customAttr'] = None

    if mod_code not in modifiers[prod_code]['parts']:
        modifiers[prod_code]['parts'][mod_code] = {
            'isOption': int(is_option) > 0,
            'isCombo': int(is_combo) > 0,
            'isCFH': int(is_cfh) > 0,
            'data': part_data
        }



from xml.etree import cElementTree as eTree
from typing import (
    Any,
    List,
    Optional,
    Union
)

from application.domain import TableService, Table, TablePictureToOrdersParser
from pos_model import TableStatus
from sysactions import get_model, get_posot, get_poslist
from tablemgrapi import get_pos_ts

from application.interactor import TableInteractor
from model.customexception import GettingUserInfoException
from utilfunctions import get_custom_property_value


class MsgBusTableService(TableService):
    def get_selected_table(self, pos_id, tables=None):
        if tables is None:
            tables = self.list_tables(pos_id)
        selected_tables = filter(lambda x: x.current_pos_id == pos_id, tables)

        selected_table = selected_tables[0] if len(selected_tables) > 0 else None
        if selected_table is not None:
            table_is_linked = selected_table.status == 5
            if table_is_linked:
                self.store_service(pos_id, selected_table.id)
                return None

        return selected_table

    def get_table_info(self, pos_id, table_id):
        pos_ts = _a_get_pos_ts(pos_id)
        return pos_ts.getTableInfo(table_id)

    def list_tables(
            self,
            pos_id  # type: str
    ):
        # type: (...) -> List[Table]

        pos_ts = _a_get_pos_ts(pos_id)

        tables_list = []
        for table in pos_ts.listTables(with_props=1):
            table_obj = Table()
            table_obj.id = table["id"]
            table_obj.name = table["name"] or table["id"]
            table_obj.current_pos_id = table["currentPOSId"]
            table_obj.status = int(table["status"])
            table_obj.slice_source_table_id = table.get("sliceSourceTableId", None)
            table_obj.slice_parent_table_id = table.get("sliceParentTableId", None)
            table_obj.tab_id = None
            if table["typeDescr"] == 'Tab' and table["properties"]:
                table_obj.tab_id = table.get("properties").get("TAB_ID")
            if table["properties"]:
                table_obj.mode = table.get("properties").get("MODE", None)
            table_obj.business_period = None
            if table["businessPeriod"]:
                table_obj.business_period = int(table["businessPeriod"].replace("-", ""))
            if table["sector"]:
                table_obj.sector = table["sector"]
            tables_list.append(table_obj)

        return tables_list

    def set_parent_table(
        self,
        pos_id,                 # type: str
        table_id,               # type: str
        parent_table_id=None    # type: Optional[str]
    ):
        # type: (...) -> None

        model = get_model(pos_id)
        pos_ts = get_pos_ts(model)
        user_id = self.get_user_id(model)
        pos_ts.setParentTable(
            user_id,
            table_id,
            parent_table_id
        )

    def set_current_order(
        self,
        pos_id,     # type: str
        order_id,   # type: str
        table_id    # type: str
    ):
        # type (...) -> None

        model = get_model(pos_id)
        pos_ts = get_pos_ts(model)
        user_id = self.get_user_id(model)

        if table_id is None:
            table_id = self.get_selected_table(pos_id).id

        tab = self._get_related_tab(pos_id, table_id, order_id)
        table_id = tab if tab else table_id

        pos_ts.setCurrentOrder(user_id, table_id, order_id)

    def _get_related_tab(
            self,
            pos_id,      # type: str
            table_id,    # type: str
            order_id     # type: str
    ):
        # type: (...) -> Optional[str]

        is_tab_mode = self.get_current_table_mode(pos_id, table_id) == 'tabs'
        tab = get_custom_property_value(pos_id, 'TABLE_ID', order_id)

        if is_tab_mode and tab.startswith('TAB'):
            table_interactor = TableInteractor(
                pos_id=pos_id,
                table_id=table_id,
                table_service=self)
            table_interactor.change_selected_table(
                pos_id=pos_id,
                selected_table_id=tab
            )
            return tab

    def void_service_order(self, pos_id, table_id, order_id):
        model = get_model(pos_id)
        pos_ts = get_pos_ts(model)
        user_id = self.get_user_id(model, False)

        pos_ts.voidServiceOrder(user_id, table_id, order_id)

    def start_service(self, pos_id, table_id="", seats="", tab_id="", mode="", source_table_id=""):
        # type: (Union[int, str], str, str, str, str, str) -> str

        model = get_model(pos_id)
        pos_ts = get_pos_ts(model)
        self.tab_id_is_valid(pos_id, tab_id, pos_ts)

        business_period = model.find("PosState").get("period")
        business_period = business_period[0:4] + "-" + business_period[4:6] + "-" + business_period[6:8]
        user_id = self.get_user_id(model)
        custom_properties = {"MODE": mode}
        if tab_id != "":
            custom_properties["TAB_ID"] = tab_id
        seats = '1' if seats is None else seats

        response = pos_ts.startService(user_id, business_period, table_id, seats, custom_properties, source_table_id)
        return response

    def tab_id_is_valid(self, pos_id, tab_id, pos_ts=None):
        if not pos_ts:
            pos_ts = get_pos_ts(get_model(pos_id))

        if tab_id != '':
            for table in pos_ts.listTables(with_props=1):
                if str(table["id"]).startswith("TAB") \
                        and int(table['status']) != TableStatus.CLOSED \
                        and table["properties"] is not None \
                        and "TAB_ID" in table["properties"] \
                        and table["properties"]["TAB_ID"] == tab_id:
                    raise Exception("TAB_ALREADY_EXISTS")

    def is_a_tab(self, pos_id, table_id, table_picture=None):
        if not table_picture:
            table_picture = eTree.XML(self.get_table_info(pos_id, table_id))

        if type(table_picture) in [str, unicode]:
            table_picture = eTree.XML(table_picture)

        if table_picture.get("id").startswith("TAB"):
            return True
        return False

    def recall_service(self, pos_id, table_id):
        model = get_model(pos_id)
        pos_ts = get_pos_ts(model)
        user_id = self.get_user_id(model)

        return pos_ts.recallService(user_id, table_id)

    def store_service(self, pos_id, table_id):
        model = get_model(pos_id)
        pos_ts = get_pos_ts(model)
        user_id = self.get_user_id(model, False)

        return pos_ts.storeService(user_id, table_id)

    def create_order(self, pos_id, table_id, price_list=''):
        model = get_model(pos_id)
        pos_ts = get_pos_ts(model)
        user_id = self.get_user_id(model)

        return pos_ts.createOrder(user_id, table_id, price_list)

    def slice_service(self, pos_id, new_setup, src_table_id):
        model = get_model(pos_id)
        pos_ts = get_pos_ts(model)
        user_id = self.get_user_id(model)

        return pos_ts.sliceService(user_id, new_setup, src_table_id)

    def get_table_picture(self, pos_id, table_id):
        pos_ts = _a_get_pos_ts(pos_id)
        return pos_ts.getTablePicture(table_id)

    def get_current_pos_id(self, pos_id, table_id):
        model = get_model(pos_id)
        table = model.find('STORE_MODIFIED/Tables/Table[@id="{}"]'.format(table_id))
        return table.get("currentPOSId")

    def do_total_table(
        self,
        pos_id,                     # type: str
        table_id,                   # type: str
        tip_rate="",                # type: Optional[str]
        selected_to_total=False     # type: Optional[bool]
    ):
        # type: (...) -> str

        table_to_total_id = table_id

        if selected_to_total:
            self.set_table_custom_property(
                pos_id,
                "SELECTED_TO_TOTAL",
                selected_to_total,
                table_id
            )
            if selected_to_total != "all_services":
                table_to_total_id = selected_to_total

        model = get_model(pos_id)
        user_id = self.get_user_id(model)

        pos_ts = get_pos_ts(model)
        return pos_ts.totalService(
            user_id=user_id,
            table_id=table_to_total_id,
            tip_rate=tip_rate
        )

    def reopen_table(self, pos_id, table_id):
        model = get_model(pos_id)
        pos_ts = get_pos_ts(model)
        user_id = self.get_user_id(model, False)

        return pos_ts.reopenService(user_id, table_id)

    def close_table(self, pos_id, table_id):
        model = get_model(pos_id)
        pos_ts = get_pos_ts(model)
        user_id = self.get_user_id(model)

        return pos_ts.closeService(user_id, table_id)

    def set_table_available(self, pos_id, table_id):
        model = get_model(pos_id)
        pos_ts = get_pos_ts(model)
        user_id = self.get_user_id(model, False)

        pos_ts.tableReady4Service(user_id, table_id)

    def get_current_table_id(self, pos_id):
        try:
            model = get_model(pos_id)
            table = model.find('STORE_MODIFIED/Tables/Table[@currentPOSId="{}"]'.format(pos_id))
            return table.get("id")
        except Exception as _: # noqa
            return None

    def clear_service_tenders(self, pos_id, table_id):
        model = get_model(pos_id)
        pos_ts = get_pos_ts(model)
        user_id = self.get_user_id(model, False)

        return pos_ts.clearServiceTenders(user_id, table_id)

    def change_service_tip(self, pos_id, table_id, percentage=None, amount=None):
        model = get_model(pos_id)
        pos_ts = get_pos_ts(model)
        user_id = self.get_user_id(model)

        return pos_ts.changeServiceTip(user_id, table_id, percentage, amount)

    def set_table_custom_property(self, pos_id, prop, prop_value=None, table_id=None, service_id=None):
        model = get_model(pos_id)
        pos_ts = get_pos_ts(model)
        pos_ts.setCustomProperty(prop, prop_value, table_id, service_id)

    def get_table_custom_property(self, pos_id, prop, table_id=None, service_id=None):
        model = get_model(pos_id)
        pos_ts = get_pos_ts(model)
        return pos_ts.getCustomProperty(prop, table_id, service_id)

    def register_service_tender(self, pos_id, table_id, tender_id, amount, tender_details=""):
        model = get_model(pos_id)
        pos_ts = get_pos_ts(model)
        user_id = self.get_user_id(model)

        return pos_ts.registerServiceTender(
            user_id,
            table_id,
            tender_id,
            str(amount),
            tender_detail=tender_details)

    def move_child_table(
        self,
        pos_id,                 # type: Any[Union[int, str]]
        source_table_id,        # type: str
        destination_table_id,   # type: str
        child_table_id,         # type: str
        mode                    # type: str
    ):

        model = get_model(pos_id)
        pos_ts = get_pos_ts(model)
        user_id = self.get_user_id(model)
        custom_properties = {"MODE": mode}
        pos_ts.moveChildTable(
            user_id=user_id,
            from_table_id=source_table_id,
            dest_table_id=destination_table_id,
            child_table_id=child_table_id,
            custom_props=custom_properties
        )

    def get_table_orders(self, pos_id, table_id, table_picture=None):
        if table_picture is None:
            table_picture = self.get_table_picture(pos_id, table_id)
        orders = TablePictureToOrdersParser.parse(table_picture)

        return orders

    def get_service_seats_count(self, table_info_xml):
        service_seats = int(table_info_xml.get("serviceSeats"))
        return int(service_seats)

    def set_sale_line_seat(self, pos_id, seat_number, order_id, line_number, item_id, level, part_code):
        self._set_order_custom_property(pos_id, 'seat', seat_number, order_id, line_number, item_id, level, part_code)

    def get_current_table_mode(
            self,
            pos_id,     # type: str
            table_id    # type: str
    ):
        # type: (...) -> str

        table_picture = eTree.XML(self.get_table_picture(
            pos_id=pos_id,
            table_id=table_id
        ))[0]
        return table_picture.get('MODE')

    @staticmethod
    def get_exclusive_pos_list(user_control_type=None, pos_list=None):
        if not pos_list:
            pos_list = get_poslist()

        pos_list = list(pos_list)

        if not user_control_type:
            return sorted(pos_list)

        for pos_id in sorted(pos_list):
            model = get_model(pos_id)
            working_mode = model.find('.//WorkingMode').get('usrCtrlType')
            if working_mode != user_control_type:
                pos_list.remove(pos_id)
        return sorted(pos_list)

    @staticmethod
    def _set_order_custom_property(pos_id, key, value, order_id, *params):
        model = get_model(pos_id)
        pos_ot = get_posot(model)

        if order_id:
            pos_ot.setOrderCustomProperty(key, value, order_id, *params)
        else:
            pos_ot.setOrderCustomProperty(key, value, '', *params)

    @staticmethod
    def get_user_id(model, mandatory=True):
        try:
            user_id = model.find("Operator").get("id")
        except Exception as ex: # noqa
            if mandatory is False:
                return '0'

            raise GettingUserInfoException(ex.args)

        return user_id


def _a_get_pos_ts(pos_id):
    model = get_model(pos_id)
    return get_pos_ts(model)



from typing import List, Dict, Any
from pos_model import Order

from _SatXmlPartBuilder import SatXmlPartBuilder


class InfCfeBuilder(SatXmlPartBuilder):
    cnf_random = ""
    cdv = ""

    def __init__(self, part_builders):
        # type: (List[SatXmlPartBuilder]) -> None
        self.part_builders = part_builders

    def build_xml(self, order, context):
        # type: (Order, Dict[unicode, Any]) -> unicode
        inner_xml = ""
        for part_builder in self.part_builders:
            inner_xml += part_builder.build_xml(order, context)

        infcfe_xml = "<infCFe versaoDadosEnt='0.07'>"
        infcfe_xml += inner_xml
        infcfe_xml += "</infCFe>"

        return infcfe_xml



# -*- coding: utf-8 -*-
from datetime import datetime
from typing import Tuple

import msgbus
import sysactions
from systools import sys_log_exception, sys_log_error

from actions.kiosk import do_void_kiosk_order
from actions.login.utils import login_user, open_user, close_user
from actions.order import create_order
from actions.util import set_pos_type
from .. import logger


@sysactions.action
def do_start_kiosk_order(pos_id, sale_type):
    try:
        model = sysactions.get_model(pos_id)

        if sysactions.has_current_order(model):
            do_void_kiosk_order(pos_id)

        user_id = authenticate_kiosk_operator(pos_id)

        check_kiosk_pos_state(pos_id, user_id)
        check_kiosk_operator_state(pos_id, user_id)
        create_kiosk_order(pos_id, sale_type)

        return True
    except Exception as _:
        sys_log_exception("Error starting kiosk order")


def create_kiosk_order(pos_id, sale_type):
    from table_actions import get_price_list_from_model
    from posactions import (
        get_pos_lock,
        new_order_properties
    )

    with get_pos_lock(pos_id=pos_id):
        model = sysactions.get_model(pos_id)
        pos_ot = sysactions.get_posot(model)
        price_list = get_price_list_from_model(pos_id, sale_type=sale_type)
        create_order(pos_ot, pos_id, price_list, sale_type=sale_type)
        set_pos_type(pos_ot, model, sale_type=sale_type)

        order_properties_dict = new_order_properties(pos_id, pos_ot)
        if len(order_properties_dict) > 0:
            pos_ot.setOrderCustomProperties(order_properties_dict)


def check_kiosk_operator_state(pos_id, user_id):
    model = sysactions.get_model(pos_id)
    operator = model.find("Operator")
    operator_state = operator.get("state") if operator is not None else None
    if not sysactions.has_operator_opened(model):
        open_kiosk_operator_automatically(pos_id, user_id, operator_state)


def check_kiosk_pos_state(pos_id, user_id):
    model = sysactions.get_model(pos_id)
    model_pos_state = model.find("PosState")
    pos_state = model_pos_state.get("state") if model_pos_state is not None else None
    pos_business_period = sysactions.get_business_period(model)
    actual_business_period = datetime.now().strftime("%Y%m%d")
    if pos_business_period != actual_business_period:
        pos_state = "BLOCKED"
    if pos_state in ["BLOCKED", "CLOSED", "UNDEFINED"]:
        open_kiosk_business_day_automatically(model, pos_id, user_id, pos_state, actual_business_period)


def open_kiosk_operator_automatically(pos_id, user_id, operator_state):
    try:
        if operator_state != "PAUSED" and not open_user(pos_id, user_id):
            raise sysactions.StopAction()
        if not login_user(pos_id, user_id):
            raise sysactions.StopAction()
    except Exception as _:
        sys_log_exception("Error on kiosk auto operator logging")
        raise


def get_kiosk_user_information(pos_id):
    # type: (str) -> Tuple[str, str]

    user_id = "2718" + str(pos_id).zfill(4)
    user_password = "2718" + str(pos_id).zfill(4)
    return user_id, user_password


def authenticate_kiosk_operator(pos_id):
    # type: (str) -> str

    user_id, user_password = get_kiosk_user_information(pos_id)
    try:
        sysactions.authenticate_user(user_id, user_password)
    except sysactions.AuthenticationFailed as _:
        sys_log_error("Error authenticating kiosk operator")
        raise sysactions.StopAction()
    return user_id


def open_kiosk_business_day_automatically(model, pos_id, user_id, pos_state, actual_business_period):
    try:
        msg_dest_name = "POS{}".format(pos_id)

        if pos_state == "BLOCKED":
            if _need_to_close_user(model) and not close_user(pos_id, user_id):
                logger.error("Need to close user to open a new business period and POS has a user logged in")
                raise sysactions.StopAction()

            if sysactions.is_day_opened(model):
                token = msgbus.TK_POS_BUSINESSEND
                msg = sysactions.send_message(msg_dest_name, token, msgbus.FM_PARAM, pos_id, 90 * 1000000)
                if msg.token == msgbus.TK_SYS_NAK:
                    logger.error("Error closing the business day on POS: {}".format(msg.data))
                    raise sysactions.StopAction()

        msg_data = "\0".join([pos_id, actual_business_period])
        token = msgbus.TK_POS_BUSINESSBEGIN
        msg = sysactions.send_message(msg_dest_name, token, msgbus.FM_PARAM, msg_data, 90 * 1000000)
        if msg.token == msgbus.TK_SYS_NAK:
            logger.error("Error starting the business day on POS: {}".format(msg.data))
            raise sysactions.StopAction()
    except Exception as _:
        logger.error("Error opening the business day on POS: {}".format(pos_id))
        sys_log_exception("Error opening business day on POS: {}".format(pos_id))
        raise


def _need_to_close_user(model):
    for operator in model.findall("Operators/Operator"):
        if operator.get("state") in ["LOGGEDIN", "OPENED", "PAUSED"]:
            return True

    return False



class PickListItem(object):
    def __init__(self, item):
        self.sub_items = []

        qty, default_qty, item_type, description, multiplied_qty = map(item.get, ("qty",
                                                                                  "default_qty",
                                                                                  "item_type",
                                                                                  "description",
                                                                                  "multiplied_qty"))

        self.qty = qty
        self.default_qty = default_qty
        self.multiplied_qty = multiplied_qty
        self.item_type = item_type
        self.description = description

        items = item.findall("Item")
        for item in items:
            self.sub_items.append(PickListItem(item))



data = (
'jjyim',    # 0x00
'jjyib',    # 0x01
'jjyibs',    # 0x02
'jjyis',    # 0x03
'jjyiss',    # 0x04
'jjying',    # 0x05
'jjyij',    # 0x06
'jjyic',    # 0x07
'jjyik',    # 0x08
'jjyit',    # 0x09
'jjyip',    # 0x0a
'jjyih',    # 0x0b
'jji',    # 0x0c
'jjig',    # 0x0d
'jjigg',    # 0x0e
'jjigs',    # 0x0f
'jjin',    # 0x10
'jjinj',    # 0x11
'jjinh',    # 0x12
'jjid',    # 0x13
'jjil',    # 0x14
'jjilg',    # 0x15
'jjilm',    # 0x16
'jjilb',    # 0x17
'jjils',    # 0x18
'jjilt',    # 0x19
'jjilp',    # 0x1a
'jjilh',    # 0x1b
'jjim',    # 0x1c
'jjib',    # 0x1d
'jjibs',    # 0x1e
'jjis',    # 0x1f
'jjiss',    # 0x20
'jjing',    # 0x21
'jjij',    # 0x22
'jjic',    # 0x23
'jjik',    # 0x24
'jjit',    # 0x25
'jjip',    # 0x26
'jjih',    # 0x27
'ca',    # 0x28
'cag',    # 0x29
'cagg',    # 0x2a
'cags',    # 0x2b
'can',    # 0x2c
'canj',    # 0x2d
'canh',    # 0x2e
'cad',    # 0x2f
'cal',    # 0x30
'calg',    # 0x31
'calm',    # 0x32
'calb',    # 0x33
'cals',    # 0x34
'calt',    # 0x35
'calp',    # 0x36
'calh',    # 0x37
'cam',    # 0x38
'cab',    # 0x39
'cabs',    # 0x3a
'cas',    # 0x3b
'cass',    # 0x3c
'cang',    # 0x3d
'caj',    # 0x3e
'cac',    # 0x3f
'cak',    # 0x40
'cat',    # 0x41
'cap',    # 0x42
'cah',    # 0x43
'cae',    # 0x44
'caeg',    # 0x45
'caegg',    # 0x46
'caegs',    # 0x47
'caen',    # 0x48
'caenj',    # 0x49
'caenh',    # 0x4a
'caed',    # 0x4b
'cael',    # 0x4c
'caelg',    # 0x4d
'caelm',    # 0x4e
'caelb',    # 0x4f
'caels',    # 0x50
'caelt',    # 0x51
'caelp',    # 0x52
'caelh',    # 0x53
'caem',    # 0x54
'caeb',    # 0x55
'caebs',    # 0x56
'caes',    # 0x57
'caess',    # 0x58
'caeng',    # 0x59
'caej',    # 0x5a
'caec',    # 0x5b
'caek',    # 0x5c
'caet',    # 0x5d
'caep',    # 0x5e
'caeh',    # 0x5f
'cya',    # 0x60
'cyag',    # 0x61
'cyagg',    # 0x62
'cyags',    # 0x63
'cyan',    # 0x64
'cyanj',    # 0x65
'cyanh',    # 0x66
'cyad',    # 0x67
'cyal',    # 0x68
'cyalg',    # 0x69
'cyalm',    # 0x6a
'cyalb',    # 0x6b
'cyals',    # 0x6c
'cyalt',    # 0x6d
'cyalp',    # 0x6e
'cyalh',    # 0x6f
'cyam',    # 0x70
'cyab',    # 0x71
'cyabs',    # 0x72
'cyas',    # 0x73
'cyass',    # 0x74
'cyang',    # 0x75
'cyaj',    # 0x76
'cyac',    # 0x77
'cyak',    # 0x78
'cyat',    # 0x79
'cyap',    # 0x7a
'cyah',    # 0x7b
'cyae',    # 0x7c
'cyaeg',    # 0x7d
'cyaegg',    # 0x7e
'cyaegs',    # 0x7f
'cyaen',    # 0x80
'cyaenj',    # 0x81
'cyaenh',    # 0x82
'cyaed',    # 0x83
'cyael',    # 0x84
'cyaelg',    # 0x85
'cyaelm',    # 0x86
'cyaelb',    # 0x87
'cyaels',    # 0x88
'cyaelt',    # 0x89
'cyaelp',    # 0x8a
'cyaelh',    # 0x8b
'cyaem',    # 0x8c
'cyaeb',    # 0x8d
'cyaebs',    # 0x8e
'cyaes',    # 0x8f
'cyaess',    # 0x90
'cyaeng',    # 0x91
'cyaej',    # 0x92
'cyaec',    # 0x93
'cyaek',    # 0x94
'cyaet',    # 0x95
'cyaep',    # 0x96
'cyaeh',    # 0x97
'ceo',    # 0x98
'ceog',    # 0x99
'ceogg',    # 0x9a
'ceogs',    # 0x9b
'ceon',    # 0x9c
'ceonj',    # 0x9d
'ceonh',    # 0x9e
'ceod',    # 0x9f
'ceol',    # 0xa0
'ceolg',    # 0xa1
'ceolm',    # 0xa2
'ceolb',    # 0xa3
'ceols',    # 0xa4
'ceolt',    # 0xa5
'ceolp',    # 0xa6
'ceolh',    # 0xa7
'ceom',    # 0xa8
'ceob',    # 0xa9
'ceobs',    # 0xaa
'ceos',    # 0xab
'ceoss',    # 0xac
'ceong',    # 0xad
'ceoj',    # 0xae
'ceoc',    # 0xaf
'ceok',    # 0xb0
'ceot',    # 0xb1
'ceop',    # 0xb2
'ceoh',    # 0xb3
'ce',    # 0xb4
'ceg',    # 0xb5
'cegg',    # 0xb6
'cegs',    # 0xb7
'cen',    # 0xb8
'cenj',    # 0xb9
'cenh',    # 0xba
'ced',    # 0xbb
'cel',    # 0xbc
'celg',    # 0xbd
'celm',    # 0xbe
'celb',    # 0xbf
'cels',    # 0xc0
'celt',    # 0xc1
'celp',    # 0xc2
'celh',    # 0xc3
'cem',    # 0xc4
'ceb',    # 0xc5
'cebs',    # 0xc6
'ces',    # 0xc7
'cess',    # 0xc8
'ceng',    # 0xc9
'cej',    # 0xca
'cec',    # 0xcb
'cek',    # 0xcc
'cet',    # 0xcd
'cep',    # 0xce
'ceh',    # 0xcf
'cyeo',    # 0xd0
'cyeog',    # 0xd1
'cyeogg',    # 0xd2
'cyeogs',    # 0xd3
'cyeon',    # 0xd4
'cyeonj',    # 0xd5
'cyeonh',    # 0xd6
'cyeod',    # 0xd7
'cyeol',    # 0xd8
'cyeolg',    # 0xd9
'cyeolm',    # 0xda
'cyeolb',    # 0xdb
'cyeols',    # 0xdc
'cyeolt',    # 0xdd
'cyeolp',    # 0xde
'cyeolh',    # 0xdf
'cyeom',    # 0xe0
'cyeob',    # 0xe1
'cyeobs',    # 0xe2
'cyeos',    # 0xe3
'cyeoss',    # 0xe4
'cyeong',    # 0xe5
'cyeoj',    # 0xe6
'cyeoc',    # 0xe7
'cyeok',    # 0xe8
'cyeot',    # 0xe9
'cyeop',    # 0xea
'cyeoh',    # 0xeb
'cye',    # 0xec
'cyeg',    # 0xed
'cyegg',    # 0xee
'cyegs',    # 0xef
'cyen',    # 0xf0
'cyenj',    # 0xf1
'cyenh',    # 0xf2
'cyed',    # 0xf3
'cyel',    # 0xf4
'cyelg',    # 0xf5
'cyelm',    # 0xf6
'cyelb',    # 0xf7
'cyels',    # 0xf8
'cyelt',    # 0xf9
'cyelp',    # 0xfa
'cyelh',    # 0xfb
'cyem',    # 0xfc
'cyeb',    # 0xfd
'cyebs',    # 0xfe
'cyes',    # 0xff
)



from ._HttpRequest import HttpRequest
from ._HttpResponse import HttpResponse
from ._HttpClient import HttpClient
from ._DefaultHttpClient import DefaultHttpClient

__all__ = ["HttpRequest", "HttpResponse", "HttpClient", "DefaultHttpClient"]



from datetime import datetime  # noqa


class LogoutSummaryInfo(object):
    def __init__(self, pos_id, store_id, operator_id, business_date, authorizer, login_time, session_id, query_date):
        # type: (int, int, int, unicode, float, unicode, unicode, unicode, unicode, datetime) -> ()
        self.pos_id = pos_id
        self.store_id = store_id
        self.operator_id = operator_id
        self.business_date = business_date
        self.authorizer = authorizer
        self.login_time = login_time
        self.session_id = session_id
        self.query_date = query_date



import json


def create_delivery_manual_order_json(pos_ot, xml_order):
    name = _get_custom_property(xml_order, "CUSTOMER_NAME")
    phone = _get_custom_property(xml_order, "CUSTOMER_PHONE")
    document = _get_custom_property(xml_order, "CUSTOMER_DOC")
    city = _get_custom_property(xml_order, "CITY")
    neighborhood = _get_custom_property(xml_order, "NEIGHBORHOOD")
    reference = _get_custom_property(xml_order, "ADDRESS_REFERENCE")
    complement = _get_custom_property(xml_order, "COMPLEMENT")
    zip_code = _get_custom_property(xml_order, "POSTAL_CODE")
    number = _get_custom_property(xml_order, "STREET_NUMBER")
    street = _get_custom_property(xml_order, "STREET_NAME")

    all_sale_lines = xml_order.findall("SaleLine")
    main_sale_lines = [x for x in all_sale_lines if x.get("level") == "0"]
    manual_delivery_order_json = {
        "items": _get_items(all_sale_lines, main_sale_lines),
        "custom_properties": [
            {
                "key": "contact_number",
                "value": phone
            }, {
                "key": "customer_doc",
                "value": document
            }, {
                "key": "customer_name",
                "value": name
            }, {
                "key": "customer_phone",
                "value": phone
            }
        ],
        "pickup": {
            "address": {
                "city": city,
                "neighborhood": neighborhood,
                "reference": reference,
                "complement": complement,
                "postalCode": zip_code,
                "formattedAddress": street + ", " + number if street and number else "",
                "streetNumber": number,
                "streetName": street
            }
        },
        "totalPrice": float(xml_order.get("totalGross", 0)),
        "partner": "Delivery",
        "shortReference": xml_order.get("orderId", "-"),
        "subTotal": float(xml_order.get("totalGross", 0)),
        "createAt": xml_order.get("createdAt", "") + "Z"
    }

    pos_ot.setOrderCustomProperty("MANUAL_DELIVERY_ORDER_JSON", json.dumps(manual_delivery_order_json))


def _get_custom_property(xml_order, custom_property):
    value = ""
    current_property = xml_order.find("CustomOrderProperties/OrderProperty[@key='{}']".format(custom_property))
    if current_property is not None:
        value = current_property.get("value", "")

    return value


def _get_items(all_sale_lines, sale_lines, level=0):
    items = []
    for sale_line in sale_lines:
        quantity = float(sale_line.get("qty", 0))
        if quantity == 0:
            continue

        line_number = sale_line.get("lineNumber")
        price = float(sale_line.get("itemPrice", 0)) / quantity
        child_sale_lines = [x for x in all_sale_lines if
                            x.get("lineNumber") == line_number and
                            x.get("level") == str(level + 1) and
                            x.get("itemId") == (sale_line.get("itemId") + "." + sale_line.get("partCode"))]
        items.append(
                {
                    "price": price,
                    "parts": _get_items(all_sale_lines, child_sale_lines, level + 1),
                    "partCode": sale_line.get("partCode", ""),
                    "observation": "",
                    "quantity": quantity if sale_line.get("itemType", "") != "OPTION" else 1,
                    "itemType": sale_line.get("itemType", "")
                }
        )

    return items



data = (
'gwan',    # 0x00
'gwanj',    # 0x01
'gwanh',    # 0x02
'gwad',    # 0x03
'gwal',    # 0x04
'gwalg',    # 0x05
'gwalm',    # 0x06
'gwalb',    # 0x07
'gwals',    # 0x08
'gwalt',    # 0x09
'gwalp',    # 0x0a
'gwalh',    # 0x0b
'gwam',    # 0x0c
'gwab',    # 0x0d
'gwabs',    # 0x0e
'gwas',    # 0x0f
'gwass',    # 0x10
'gwang',    # 0x11
'gwaj',    # 0x12
'gwac',    # 0x13
'gwak',    # 0x14
'gwat',    # 0x15
'gwap',    # 0x16
'gwah',    # 0x17
'gwae',    # 0x18
'gwaeg',    # 0x19
'gwaegg',    # 0x1a
'gwaegs',    # 0x1b
'gwaen',    # 0x1c
'gwaenj',    # 0x1d
'gwaenh',    # 0x1e
'gwaed',    # 0x1f
'gwael',    # 0x20
'gwaelg',    # 0x21
'gwaelm',    # 0x22
'gwaelb',    # 0x23
'gwaels',    # 0x24
'gwaelt',    # 0x25
'gwaelp',    # 0x26
'gwaelh',    # 0x27
'gwaem',    # 0x28
'gwaeb',    # 0x29
'gwaebs',    # 0x2a
'gwaes',    # 0x2b
'gwaess',    # 0x2c
'gwaeng',    # 0x2d
'gwaej',    # 0x2e
'gwaec',    # 0x2f
'gwaek',    # 0x30
'gwaet',    # 0x31
'gwaep',    # 0x32
'gwaeh',    # 0x33
'goe',    # 0x34
'goeg',    # 0x35
'goegg',    # 0x36
'goegs',    # 0x37
'goen',    # 0x38
'goenj',    # 0x39
'goenh',    # 0x3a
'goed',    # 0x3b
'goel',    # 0x3c
'goelg',    # 0x3d
'goelm',    # 0x3e
'goelb',    # 0x3f
'goels',    # 0x40
'goelt',    # 0x41
'goelp',    # 0x42
'goelh',    # 0x43
'goem',    # 0x44
'goeb',    # 0x45
'goebs',    # 0x46
'goes',    # 0x47
'goess',    # 0x48
'goeng',    # 0x49
'goej',    # 0x4a
'goec',    # 0x4b
'goek',    # 0x4c
'goet',    # 0x4d
'goep',    # 0x4e
'goeh',    # 0x4f
'gyo',    # 0x50
'gyog',    # 0x51
'gyogg',    # 0x52
'gyogs',    # 0x53
'gyon',    # 0x54
'gyonj',    # 0x55
'gyonh',    # 0x56
'gyod',    # 0x57
'gyol',    # 0x58
'gyolg',    # 0x59
'gyolm',    # 0x5a
'gyolb',    # 0x5b
'gyols',    # 0x5c
'gyolt',    # 0x5d
'gyolp',    # 0x5e
'gyolh',    # 0x5f
'gyom',    # 0x60
'gyob',    # 0x61
'gyobs',    # 0x62
'gyos',    # 0x63
'gyoss',    # 0x64
'gyong',    # 0x65
'gyoj',    # 0x66
'gyoc',    # 0x67
'gyok',    # 0x68
'gyot',    # 0x69
'gyop',    # 0x6a
'gyoh',    # 0x6b
'gu',    # 0x6c
'gug',    # 0x6d
'gugg',    # 0x6e
'gugs',    # 0x6f
'gun',    # 0x70
'gunj',    # 0x71
'gunh',    # 0x72
'gud',    # 0x73
'gul',    # 0x74
'gulg',    # 0x75
'gulm',    # 0x76
'gulb',    # 0x77
'guls',    # 0x78
'gult',    # 0x79
'gulp',    # 0x7a
'gulh',    # 0x7b
'gum',    # 0x7c
'gub',    # 0x7d
'gubs',    # 0x7e
'gus',    # 0x7f
'guss',    # 0x80
'gung',    # 0x81
'guj',    # 0x82
'guc',    # 0x83
'guk',    # 0x84
'gut',    # 0x85
'gup',    # 0x86
'guh',    # 0x87
'gweo',    # 0x88
'gweog',    # 0x89
'gweogg',    # 0x8a
'gweogs',    # 0x8b
'gweon',    # 0x8c
'gweonj',    # 0x8d
'gweonh',    # 0x8e
'gweod',    # 0x8f
'gweol',    # 0x90
'gweolg',    # 0x91
'gweolm',    # 0x92
'gweolb',    # 0x93
'gweols',    # 0x94
'gweolt',    # 0x95
'gweolp',    # 0x96
'gweolh',    # 0x97
'gweom',    # 0x98
'gweob',    # 0x99
'gweobs',    # 0x9a
'gweos',    # 0x9b
'gweoss',    # 0x9c
'gweong',    # 0x9d
'gweoj',    # 0x9e
'gweoc',    # 0x9f
'gweok',    # 0xa0
'gweot',    # 0xa1
'gweop',    # 0xa2
'gweoh',    # 0xa3
'gwe',    # 0xa4
'gweg',    # 0xa5
'gwegg',    # 0xa6
'gwegs',    # 0xa7
'gwen',    # 0xa8
'gwenj',    # 0xa9
'gwenh',    # 0xaa
'gwed',    # 0xab
'gwel',    # 0xac
'gwelg',    # 0xad
'gwelm',    # 0xae
'gwelb',    # 0xaf
'gwels',    # 0xb0
'gwelt',    # 0xb1
'gwelp',    # 0xb2
'gwelh',    # 0xb3
'gwem',    # 0xb4
'gweb',    # 0xb5
'gwebs',    # 0xb6
'gwes',    # 0xb7
'gwess',    # 0xb8
'gweng',    # 0xb9
'gwej',    # 0xba
'gwec',    # 0xbb
'gwek',    # 0xbc
'gwet',    # 0xbd
'gwep',    # 0xbe
'gweh',    # 0xbf
'gwi',    # 0xc0
'gwig',    # 0xc1
'gwigg',    # 0xc2
'gwigs',    # 0xc3
'gwin',    # 0xc4
'gwinj',    # 0xc5
'gwinh',    # 0xc6
'gwid',    # 0xc7
'gwil',    # 0xc8
'gwilg',    # 0xc9
'gwilm',    # 0xca
'gwilb',    # 0xcb
'gwils',    # 0xcc
'gwilt',    # 0xcd
'gwilp',    # 0xce
'gwilh',    # 0xcf
'gwim',    # 0xd0
'gwib',    # 0xd1
'gwibs',    # 0xd2
'gwis',    # 0xd3
'gwiss',    # 0xd4
'gwing',    # 0xd5
'gwij',    # 0xd6
'gwic',    # 0xd7
'gwik',    # 0xd8
'gwit',    # 0xd9
'gwip',    # 0xda
'gwih',    # 0xdb
'gyu',    # 0xdc
'gyug',    # 0xdd
'gyugg',    # 0xde
'gyugs',    # 0xdf
'gyun',    # 0xe0
'gyunj',    # 0xe1
'gyunh',    # 0xe2
'gyud',    # 0xe3
'gyul',    # 0xe4
'gyulg',    # 0xe5
'gyulm',    # 0xe6
'gyulb',    # 0xe7
'gyuls',    # 0xe8
'gyult',    # 0xe9
'gyulp',    # 0xea
'gyulh',    # 0xeb
'gyum',    # 0xec
'gyub',    # 0xed
'gyubs',    # 0xee
'gyus',    # 0xef
'gyuss',    # 0xf0
'gyung',    # 0xf1
'gyuj',    # 0xf2
'gyuc',    # 0xf3
'gyuk',    # 0xf4
'gyut',    # 0xf5
'gyup',    # 0xf6
'gyuh',    # 0xf7
'geu',    # 0xf8
'geug',    # 0xf9
'geugg',    # 0xfa
'geugs',    # 0xfb
'geun',    # 0xfc
'geunj',    # 0xfd
'geunh',    # 0xfe
'geud',    # 0xff
)



# -*- coding: utf-8 -*-

import sysactions
from bustoken import TK_POS_UPDATER_PERFORM_USER_UPDATE
from msgbus import TK_SYS_NAK, FM_PARAM

from actions.update import show_nak_message
from actions.util.list_users import list_ids_opened_users
from utilfunctions import json_serialize


@sysactions.action
def update_users(pos_id):
    data = json_serialize(list_ids_opened_users())
    msg = sysactions.send_message("PosUpdater", TK_POS_UPDATER_PERFORM_USER_UPDATE, FM_PARAM, data)
    if msg.token == TK_SYS_NAK:
        response_message = msg.data
        show_nak_message(pos_id, "$UPDATE_USERS_", response_message)

    sysactions.show_messagebox(pos_id, "$UPDATE_USERS_MESSAGE", buttons="OK")



from _CloseDaySummaryHeaderDto import CloseDaySummaryHeaderDto  # noqa
from _CloseDaySummaryBodyDto import CloseDaySummaryBodyDto  # noqa


class CloseDaySummaryDto(object):
    def __init__(self, close_day_summary_header_dto, close_day_summary_body_dto):
        # type:(CloseDaySummaryHeaderDto, CloseDaySummaryBodyDto) -> ()
        self.close_day_summary_header_dto = close_day_summary_header_dto
        self.close_day_summary_body_dto = close_day_summary_body_dto



from ._MessageBusExceptionType import MessageBusExceptionType


class MessageBusException(Exception):
    def __init__(self, type):
        # type: (MessageBusExceptionType) -> None
        self.type = type



from abc import ABCMeta, abstractmethod
from logging import Logger

from cfgtools import Configuration
from dialogmanager import DialogManager
from messagebus import MessageBus, Message, DataType
from modelcommon.model import SitefResponse
from typing import (
    Any,
    Union
)

from model import (
    CancelPaymentOrderData,
    ConfirmFinishPaymentData,
    EftData,
    FinishPaymentData,
    PaymentResponse,
    StartPaymentData,
    TestConnectionData
)

TK_ACCOUNT_EFTACTIVITY = 0x00E00005


class PaymentProcessor(object):
    __metaclass__ = ABCMeta

    def __init__(self, message_bus, dialog_manager, config, logger):
        # type: (MessageBus, DialogManager, Configuration, Logger) -> None

        self.message_bus = message_bus
        self.dialog_manager = dialog_manager
        self.config = config
        self.logger = logger

    @abstractmethod
    def connection_status(self, test_connection_data):
        # type: (TestConnectionData) -> bool
        raise NotImplementedError()

    @abstractmethod
    def add_payment(self, start_payment_data):
        # type: (StartPaymentData) -> PaymentResponse
        raise NotImplementedError()

    @abstractmethod
    def cancel_payment_order(
        self,
        cancel_payment_order_data   # type: CancelPaymentOrderData
    ):
        # type: (...) -> Any[Union[SitefResponse, str]]

        raise NotImplementedError()

    @abstractmethod
    def cancel_payment(self):
        # type: () -> str
        raise NotImplementedError()

    @abstractmethod
    def finish_payment(
        self,
        finish_payment_data     # type: FinishPaymentData
    ):
        # type: (...) -> bool
        raise NotImplementedError()

    @abstractmethod
    def confirm_finish_payment(
        self,
        confirm_finish_payment_data     # type: ConfirmFinishPaymentData
    ):
        # type: (...) -> None
        raise NotImplementedError()

    @abstractmethod
    def get_name(self):
        # type: () -> str
        raise NotImplementedError()

    def stop(self):
        # type: () -> None
        pass

    def register_eft_activity(self, eft_data):
        # type: (EftData) -> None
        self.message_bus.send_message("account{}".format(eft_data.pos_id),
                                      Message(TK_ACCOUNT_EFTACTIVITY,
                                              DataType.param, eft_data.to_payload()))



# -*- coding: utf-8 -*-

import logging
import logging.config
import logging.handlers
import os
import re
import sqlite3
import sys
import time
from StringIO import StringIO
from abc import ABCMeta
from datetime import datetime, timedelta
from decimal import ROUND_HALF_UP, Decimal
from multiprocessing.pool import ThreadPool
from xml.etree import cElementTree as eTree

import six
from dateutil import tz
from msgbus import FM_PARAM, TK_SYS_NAK, TK_STORECFG_GET, TK_POS_GETPOSLIST, MBEasyContext
from persistence import Driver, Connection, DbdException
from typing import Callable, Any, List, Optional, Union


def F(value):
    if value is None:
        return 0
    return float(value)
# END F


def get_sale_line_priced_items(sale_lines, ignore_price=False):
    deleted_line_numbers = map(lambda x: x.get("lineNumber"), filter(lambda x: x.get("level") == "0" and x.get("qty") == "0", sale_lines))
    if ignore_price:
        active_sale_lines = filter(lambda x: (x.get("lineNumber") not in deleted_line_numbers), sale_lines)
    else:
        active_sale_lines = filter(lambda x: (x.get("lineNumber") not in deleted_line_numbers) and (F(x.get("itemPrice")) != 0), sale_lines)

    for sale_line in active_sale_lines:
        item_price, unit_price, item_qty, added_price, added_qty, multiplied_qty, inc_qty, level = map(sale_line.get, (("itemPrice", "unitPrice", "qty", "addedUnitPrice", "addedQty", "multipliedQty", "incQty", "level")))
        unit_price_cache = F(sale_line.get("addedUnitPrice")) or F(sale_line.get("unitPrice"))
        try:
            if F(unit_price) > 0:
                price_quantity = round_half_away_from_zero(F(item_price) / F(unit_price), 2)
                product_quantity = round_half_away_from_zero(F(item_qty) * F(multiplied_qty), 2)
                if price_quantity != product_quantity:
                    unit_price_cache = F(item_price) / product_quantity
                    unit_price = unit_price_cache
        except ZeroDivisionError:
            unit_price = 0
        try:
            correct_price = (F(unit_price) * F(item_qty) + F(added_price) * F(added_qty or item_qty)) * (F(multiplied_qty) / F(item_qty))

        except ZeroDivisionError:
            correct_price = 0

        unit_price = unit_price_cache
        if unit_price == 0 and F(item_price) > 0:
            correct_price = F(item_price)
        if not ignore_price:
            item_price = correct_price
            sale_line.set("correctItemPrice", str(item_price))

    return active_sale_lines
# END get_sale_line_priced_items


def read_swconfig(mbcontext, key):
    rmsg = mbcontext.MB_EasySendMessage("StoreWideConfig", token=TK_STORECFG_GET, format=FM_PARAM, data=key)
    if rmsg.token == TK_SYS_NAK:
        return None

    return str(rmsg.data)
# END read_swconfig


def get_log_file_path(loader_path, log_name='log.conf'):
    dir_path = os.path.dirname(os.path.realpath(loader_path))
    return os.path.join(dir_path, log_name)
# END get_log_file_path


def round_half_away_from_zero(number, precision=0):
    number = Decimal(number)
    value = number.quantize(Decimal('10') ** -(precision + 2))
    return float(value.quantize(Decimal('10') ** -precision, rounding=ROUND_HALF_UP))


def config_logger(loader_path, log_name='component'):
    dir_path = os.path.dirname(os.path.realpath(loader_path))
    filename = os.path.join(dir_path, log_name + '.log')

    formatter = logging.Formatter('%(asctime)-6s: %(name)s - %(levelname)s - %(thread)6d - %(threadName)-12s - %(message)s')

    rotating_file_logger = logging.handlers.RotatingFileHandler(filename, 'a', 5242880, 5)
    rotating_file_logger.setLevel(logging.DEBUG)  # DEBUG Log
    # rotating_file_logger.setLevel(logging.ERROR)  # Production Log
    rotating_file_logger.setFormatter(formatter)

    comp_logger = logging.getLogger(log_name)
    comp_logger.addHandler(rotating_file_logger)
    comp_logger.setLevel(logging.NOTSET)

    console_logger = logging.StreamHandler(sys.stdout)
    console_logger.setLevel(logging.ERROR)
    console_logger.setFormatter(formatter)

    root_logger = logging.getLogger()
    root_logger.addHandler(console_logger)
    root_logger.setLevel(logging.NOTSET)
# END get_log_file_path


class OrderTaker(object):
    def __init__(self):
        return

    def get_order_picture(self, pos_id, order_id=None):
        from sysactions import get_model, get_posot
        model = get_model(pos_id)
        posot = get_posot(model)
        if order_id is None:
            order = posot.orderPicture(pos_id)
            orders_xml = eTree.XML(order)

            return eTree.tostring(orders_xml)
        else:
            orders = posot.orderPicture(orderid=order_id)
            orders_xml = eTree.XML(orders)
            order = orders_xml.find("Order")

            return eTree.tostring(order)


class PosUtil(object):
    def __init__(self, mbcontext):
        self.mbcontext = mbcontext

    def get_pos_list(self):
        msg = self.mbcontext.MB_EasySendMessage("PosController", TK_POS_GETPOSLIST)
        if msg.token == TK_SYS_NAK:
            raise Exception("Could not retrieve PosList")

        pos_list = sorted(map(int, msg.data.split("\0")))

        return pos_list


def convert_from_utf_to_localtime(utc):
    # type: (datetime) -> datetime
    from_zone = tz.tzutc()
    to_zone = tz.tzlocal()

    # Tell the datetime object that it's in UTC time zone since
    # datetime objects are 'naive' by default
    utc = utc.replace(tzinfo=from_zone)

    # Convert time zone
    localtime = utc.astimezone(to_zone)

    return localtime


def convert_from_localtime_to_utc(localtime):
    # type: (datetime) -> datetime
    from_zone = tz.tzlocal()
    to_zone = tz.tzutc()

    # Tell the datetime object that it's in UTC time zone since
    # datetime objects are 'naive' by default
    localtime = localtime.replace(tzinfo=from_zone)

    # Convert time zone
    utc = localtime.astimezone(to_zone)

    return utc


def validar_cpf(cpf):
    """
    Valida CPFs, retornando apenas a string de números válida.

    # CPFs errados
    >>> validar_cpf('abcdefghijk')
    False
    >>> validar_cpf('123')
    False
    >>> validar_cpf('')
    False
    >>> validar_cpf(None)
    False
    >>> validar_cpf('12345678900')
    False

    # CPFs corretos
    >>> validar_cpf('95524361503')
    '95524361503'
    >>> validar_cpf('955.243.615-03')
    '95524361503'
    >>> validar_cpf('  955 243 615 03  ')
    '95524361503'
    """
    cpf = ''.join(re.findall('\d', str(cpf)))

    if (not cpf) or (len(cpf) < 11):
        return False

    # Pega apenas os 9 primeiros dígitos do CPF e gera os 2 dígitos que faltam
    inteiros = map(int, cpf)
    novo = inteiros[:9]

    while len(novo) < 11:
        r = sum([(len(novo) + 1 - i) * v for i, v in enumerate(novo)]) % 11

        if r > 1:
            f = 11 - r
        else:
            f = 0
        novo.append(f)

    # Se o número gerado coincidir com o número original, é válido
    if novo == inteiros:
        return cpf

    return False


def validar_cnpj(cnpj):
    """
    Valida CNPJs, retornando apenas a string de números válida.

    # CNPJs errados
    >>> validar_cnpj('abcdefghijklmn')
    False
    >>> validar_cnpj('123')
    False
    >>> validar_cnpj('')
    False
    >>> validar_cnpj(None)
    False
    >>> validar_cnpj('12345678901234')
    False
    >>> validar_cnpj('11222333000100')
    False

    # CNPJs corretos
    >>> validar_cnpj('11222333000181')
    '11222333000181'
    >>> validar_cnpj('11.222.333/0001-81')
    '11222333000181'
    >>> validar_cnpj('  11 222 333 0001 81  ')
    '11222333000181'
    """
    cnpj = ''.join(re.findall('\d', str(cnpj)))

    if (not cnpj) or (len(cnpj) < 14):
        return False

    # Pega apenas os 12 primeiros dígitos do CNPJ e gera os 2 dígitos que faltam
    inteiros = map(int, cnpj)
    novo = inteiros[:12]

    prod = [5, 4, 3, 2, 9, 8, 7, 6, 5, 4, 3, 2]
    while len(novo) < 14:
        r = sum([x * y for (x, y) in zip(novo, prod)]) % 11
        if r > 1:
            f = 11 - r
        else:
            f = 0
        novo.append(f)
        prod.insert(0, 6)

    # Se o número gerado coincidir com o número original, é válido
    if novo == inteiros:
        return cnpj

    return False


def regex_search(re_search, text, key=None):
    # type: (str, unicode, Optional[str]) -> Union[str, None]
    if key:
        re_search = re_search.format(key)
    result = re.search(re_search, text)
    if result not in (None, ""):
        return result.group(1)
    else:
        return None


def remove_xml_namespace(xml):
    # type: (Union[str, unicode]) -> eTree
    it = eTree.iterparse(StringIO(xml))
    for _, el in it:
        if '}' in el.tag:
            el.tag = el.tag.split('}', 1)[1]
    return it.root


def remove_punctuation(text):
    return re.sub('[^A-Za-z0-9 ]+', '', text)


class Clock(object):
    def get_current_date(self):
        # type: () -> datetime
        pass

    def get_current_time(self):
        # type: () -> datetime
        pass


class RealClock(Clock):
    def get_current_time(self):
        return datetime.today()

    def get_current_date(self):
        return datetime.now().date()


class BaseRepository(object):
    __metaclass__ = ABCMeta

    def __init__(self, mbcontext):
        # type: (MBEasyContext) -> None
        self.mbcontext = mbcontext

    def execute_with_connection(self, method, db_name=None, service_name=None):
        # type: (Callable[[Connection], Any], str, str) -> Any
        conn = None
        try:
            conn = Driver().open(self.mbcontext, dbname=str(db_name) if db_name is not None else None, service_name=service_name)
            return method(conn)
        finally:
            if conn is not None:
                conn.close()

    def execute_with_transaction(self, method, db_name=None, service_name=None):
        # type: (Callable[[Connection], Any], str, str) -> Any
        def transaction_method(conn):
            # type: (Connection) -> Any

            if conn.get_number_of_instances() != 1 and db_name is None:
                raise DbdException("Cannot open transaction without DB instante")

            conn.transaction_start()
            try:
                conn.query("BEGIN TRANSACTION")
                ret = method(conn)
                conn.query("COMMIT")
                return ret
            except:
                conn.query("ROLLBACK")
                raise
            finally:
                conn.transaction_end()

        return self.execute_with_connection(transaction_method, db_name=db_name, service_name=service_name)

    def execute_in_all_databases(self, method, pos_list):
        # type: (Callable[[Connection], Any], List[int]) -> Any
        ret = []
        for pos_id in pos_list:
            ret.append(self._execute_in_database(method, pos_id))

        return ret

    def execute_in_all_databases_returning_flat_list(self, method, pos_list):
        # type: (Callable[[Connection], Any], List[int]) -> Any
        inner_ret = self.execute_in_all_databases(method, pos_list)

        ret = []
        for inner_list in inner_ret:
            ret.extend(inner_list)

        return ret

    def execute_parallel_in_all_databases_returning_flat_list(self, method, pos_list):
        # type: (Callable[[Connection], Any], List[int]) -> Any
        inner_ret = self.execute_parallel_in_all_databases(method, pos_list)

        ret = []
        for inner_list in inner_ret:
            ret.extend(inner_list)

        return ret

    def execute_parallel_in_all_databases(self, method, pos_list):
        # type: (Callable[[Connection], Any], List[int]) -> Any
        ret = []
        pool = ThreadPool(len(pos_list))

        for pos_id in pos_list:
            ret.append(pool.apply_async(self._execute_in_database, args=(method, pos_id)))

        pool.close()
        pool.join()

        result = [r.get() for r in ret]
        return result

    def _execute_in_database(self, method, pos_id):
        conn = None
        try:
            conn = Driver().open(self.mbcontext, dbname=str(pos_id))
            return method(conn)
        finally:
            if conn is not None:
                conn.close()


class SQLiteBaseRepository(object):
    __metaclass__ = ABCMeta

    def __init__(self, db_path):
        # type: (unicode) -> None
        self.db_path = db_path

    def execute_with_connection(self, method):
        # type: (Callable[[sqlite3.Cursor], Any]) -> Any
        conn = None
        cursor = None
        try:
            conn = sqlite3.connect(self.db_path)
            cursor = conn.cursor()
            return method(cursor)
        finally:
            if cursor is not None:
                cursor.close()
            if conn is not None:
                conn.close()

    def execute_with_transaction(self, method):
        # type: (Callable[[sqlite3.Cursor], Any]) -> Any
        conn = None
        cursor = None
        try:
            conn = sqlite3.connect(self.db_path)
            cursor = conn.cursor()
            ret = method(cursor)
            conn.commit()
            return ret
        except:
            conn.rollback()
            six.reraise(*sys.exc_info())
        finally:
            if cursor is not None:
                cursor.close()
            if conn is not None:
                conn.close()


class ValidationException(Exception):
    def __init__(self, error_code, localized_error_message):
        # type: (unicode, unicode) -> None
        self.error_code = error_code
        self.localized_error_message = localized_error_message

    def __str__(self):
        # type: () -> unicode
        return unicode(str(type(self))) + self.error_code + u": " + self.localized_error_message

    def __repr__(self):
        # type: () -> unicode
        return self.__str__()


class InvalidJsonException(Exception):
    def __init__(self, message, invalid_json=None):
        # type: (unicode, unicode) -> None
        self.message = message
        self.invalid_json = invalid_json

    def __str__(self):
        # type: () -> unicode
        return u"InvalidJsonException: " + self.message + u" " + self.invalid_json

    def __repr__(self):
        # type: () -> unicode
        return self.__str__()


def get_date_difference_GMT_timezone(timestamp):
    date_sangria = datetime.strptime(timestamp, "%Y-%m-%d %H:%M:%S")
    date_sangria = date_sangria.replace(tzinfo=tz.tzlocal())
    adjust_hours = int(date_sangria.strftime("%z")[2:][:1])
    date_sangria = date_sangria - timedelta(hours=adjust_hours)

    return date_sangria.strftime("%Y-%m-%dT%H:%M:%S%z")


def get_class_by_name(kls):
    # type: (unicode) -> type
    parts = kls.split(".")
    module = ".".join(parts[:-1])
    m = __import__(module)
    for comp in parts[1:]:
        m = getattr(m, comp)
    return m



class ProductSaleItem(object):

    def __init__(self, name, quantity, total_price):
        # type: (str, float, float) -> None

        self.name = name
        self.quantity = quantity
        self.total_price = total_price



from report.command import FontCommand

from .._CommandProcessor import CommandProcessor


class FontCommandProcessor(CommandProcessor):
    def get_bytes(self, command):
        if not isinstance(command, FontCommand):
            return ""

        esc = b'\x1b'
        _set_font = lambda n: esc + b'\x4d' + n

        if command.font == FontCommand.Font.A:
            ret = _set_font(b'\x00')
        elif command.font == FontCommand.Font.B:
            ret = _set_font(b'\x01')
        else:
            return ""

        return ret



# -*- coding: utf-8 -*-

import glob
import logging
import os
import sqlite3
import subprocess

from datetime import datetime
from typing import (
    Dict,
    List,
    Optional,
    Tuple
)

from msgbus import MBEasyContext
from persistence import Driver
from repository import (
    FiscalRepository,
    OrderRepository
)
from services import CommonService

logger = logging.getLogger("Maintenance")


class CleanDatabasesProcessor(object):
    def __init__(
        self,
        mb_context,                 # type: MBEasyContext
        comp_no,                    # type: int
        database_backups_to_keep,   # type: int
        order_repository,           # type: OrderRepository
        fiscal_repository,          # type: FiscalRepository
        common_service              # type: CommonService
    ):
        # type: (...) -> None

        self.mb_context = mb_context
        self.database_backups_to_keep = database_backups_to_keep
        self.order_repository = order_repository
        self.fiscal_repository = fiscal_repository
        self.common_service = common_service
        self.database_path = order_repository.get_database_dir()
        self.backup_database_path = order_repository.get_backup_database_dir()

        self.component_is_server = comp_no == 0
        self.maintenance_processor = None
        self.pos_list = None

    def set_server_parameters(self, maintenance_processor):
        self.maintenance_processor = maintenance_processor

    def clean_databases(self):
        # type: () -> None

        try:
            if not self.component_is_server:
                return

            self.pos_list = self.common_service.get_pos_list()
            self._validate_backup_path()
            self._clean_orders_database()
            self._clean_fiscal_database()
        except Exception as _:
            logger.exception("Error cleaning databases")

    def vacuum_databases(self):
        # type: () -> None

        try:
            if not self.component_is_server:
                return

            self._vacuum_production_database()
            self._vacuum_fiscal_database()
            self._vacuum_orders_database()
        except Exception as _:
            logger.exception("Error doing vacuum on databases")

    def clean_database_backups(self):
        # type: () -> None

        logger.info("Starting [_clean_database_backups]")
        try:
            if not self.component_is_server:
                return

            files_to_delete = self._get_backup_files_to_delete()
            self.common_service.delete_files(files_to_delete)
        except Exception as _:
            logger.exception("Error cleaning databases backups")

        logger.info("Finished [_clean_database_backups]")

    def _validate_backup_path(self):
        # type: () -> None
        self.common_service.make_sure_path_exists(self.backup_database_path)

    def _clean_orders_database(self):
        # type: () -> None

        logger.info("Starting [_clean_orders_database]")
        try:
            for pos_id in sorted(self.pos_list, key=lambda _: int(_)):
                logger.info('Starting backup process for POS: %s' % pos_id)
                try:
                    end_date = self.common_service.get_end_date()
                    orders_stored = self._backup_orders(pos_id, end_date)

                    recalled_orders = self.order_repository.get_orders(
                        pos_id=pos_id,
                        end_date=end_date,
                        order_states=['RECALLED']
                    )
                    recalled_orders = [o[0] for o in recalled_orders]

                    orders_to_delete = orders_stored
                    orders_to_delete.extend(x for x in recalled_orders if x not in orders_to_delete)
                    if orders_to_delete:
                        self._delete_orders(
                            pos_id=pos_id,
                            orders=orders_to_delete
                        )
                    else:
                        logger.info("Have no orders to clean in order db: {}".format(pos_id))
                except Exception as _:
                    logger.exception("Error doing backup - POS: %s" % pos_id)
                    break

            logger.info('Finished [_clean_orders_database]')
        except Exception as _:
            logger.exception("Error cleaning order dbs")

    def _delete_orders(
        self,
        pos_id,     # type: int
        orders      # type: List[int]
    ):
        # type: (...) -> None

        orders_batches = [orders[x:x + 100] for x in xrange(0, len(orders), 100)]
        logger.info("Starting [_delete_orders] por POS: {}".format(pos_id))
        for orders_batch in orders_batches:
            self.order_repository.delete_orders(
                pos_id=pos_id,
                orders=orders_batch
            )
        logger.info("Finished [_delete_orders] por POS: {}".format(pos_id))

    def _vacuum_orders_database(self):
        # For each POS - Backup Orders DB
        logger.info("Starting [_vacuum_orders_database]")
        try:
            order_db_files = []
            conn = None

            for pos_id in sorted(self.pos_list, key=lambda _: int(_)):
                try:
                    conn = Driver().open(
                        mbcontext=self.mb_context,
                        service_name='Persistence',
                        dbname=str(pos_id)
                    )

                    order_db_files.append(self.get_db_file(conn, "orderdb"))
                finally:
                    if conn:
                        conn.close()

            for order_db in order_db_files:
                conn = None
                old_file_size = self.common_service.get_file_size(order_db)
                logger.info('Doing VACUUM for order db: {}'.format(order_db.split('db')[1]))

                try:
                    conn = sqlite3.connect(order_db)
                    conn.execute("VACUUM;")  # This will decrease the size of the databases
                except Exception as ex:
                    logger.exception("Error doing vacuum - %s - Message %s" % (order_db, str(ex)))
                    if conn is not None:
                        conn.close()
                else:
                    conn.close()

                new_file_size = self.common_service.get_file_size(order_db)

                logger.info('Order db file sizes. OLD: {}mb; NEW: {}mb'.format(old_file_size, new_file_size))

            logger.info('Finished [_vacuum_orders_database]')
        except Exception as _:
            logger.exception("Error doing vacuum on order dbs")

    def _get_backup_files_to_delete(self):
        # type: () -> List[str]

        files_to_delete = []

        files = glob.glob("{}/*.bkp".format(self.database_path))
        files_by_database_name = {}

        for f in files:
            created_time = os.path.getctime(f)
            file_name = os.path.basename(f)
            database_name = file_name.split('.')[0] if '.' in file_name else ''
            if database_name in files_by_database_name:
                files_by_database_name[database_name].append((f, created_time))
            else:
                files_by_database_name[database_name] = [(f, created_time)]

        for key in files_by_database_name:
            files_by_database_name[key].sort(key=lambda i: i[1], reverse=True)
            files_to_delete.extend([f[0] for f in files_by_database_name[key][self.database_backups_to_keep:]])

        return files_to_delete

    @staticmethod
    def _split_by_months(
        items,              # type: List[Tuple[int, str]]
        timestamp=False     # type: Optional[bool]
    ):
        # type: (...) -> Dict[str, List[int]]

        items_by_months = {}
        for item, day in items:
            if timestamp:
                date = datetime.fromtimestamp(int(day))
                month = date.strftime("%Y%m")
            else:
                month = day[:-2]

            if month in items_by_months:
                items_by_months[month].append(item)
            else:
                items_by_months[month] = [item]

        return items_by_months

    def _backup_orders(
        self,
        pos_id,     # type: int
        end_date,   # str
    ):
        # type: (...) -> List[int]

        logger.info("Starting [_backup_orders]")

        orders = self.order_repository.get_orders(
            pos_id=pos_id,
            end_date=end_date,
            order_states=['PAID', 'VOIDED', 'ABANDONED', 'SYSTEM_VOIDED']
        )
        if len(orders) == 0:
            return []

        orders_stored = []
        orders_by_month = self._split_by_months(orders)

        try:
            for month in orders_by_month:
                self.order_repository.fill_order_backup_database(
                    pos_id=pos_id,
                    suffix_date=month,
                    orders=orders_by_month[month]
                )

            orders_stored = [o[0] for o in orders]
            logger.info("Finished [_backup_orders]")

        except Exception as _:
            logger.exception("Error doing backup")

        return orders_stored

    def _clean_fiscal_database(self):
        try:
            logger.info("Starting [_clean_fiscal_database]")

            orders_backup = self._backup_fiscal()
            if orders_backup:
                self._delete_fiscal(orders_backup)
            else:
                logger.info("Have no order to clean in fiscal db")
        except Exception as ex:
            logger.exception('Error purging orders on fiscal database')
            raise ex
        else:
            logger.info('Finalized purge of orders on fiscal database')
            logger.info("Finished [_clean_fiscal_database]")

    def _vacuum_fiscal_database(self):
        try:
            logger.info("Starting [_vacuum_fiscal_database]")

            conn = None
            try:
                conn = Driver().open(self.mb_context, service_name="FiscalPersistence")
                db_file = self.get_db_file(conn, "fiscal")
            except Exception as ex:
                if "malformed" in ex.message:
                    logger.exception("Malformed database. Executing hard clean")
                    db_file = self.get_db_file(conn, "fiscal")
                    self._hard_clean_fiscal_database(db_file)
                logger.exception('Error doing vacuum on fiscal database')
                raise ex
            finally:
                if conn:
                    conn.close()

            old_file_size = self.common_service.get_file_size(db_file)
            conn = None
            try:
                logger.info('Doing VACUUM for fiscal database')
                if db_file:
                    conn = sqlite3.connect(db_file)
                    conn.execute("VACUUM;")
            except Exception as _:
                logger.exception('Error doing vacuum on fiscal db')
            finally:
                if conn:
                    conn.close()

            new_file_size = self.common_service.get_file_size(db_file)

            logger.info('Fiscal db file sizes. OLD: {}mb; NEW: {}mb'.format(old_file_size, new_file_size))

            logger.info("Finished [_vacuum_fiscal_database]")
        except Exception as _:
            logger.exception("Error doing vacuum on fiscal db")

    def _vacuum_production_database(self):
        try:
            logger.info("Starting [_vacuum_production_database]")

            conn = None
            try:
                conn = Driver().open(self.mb_context)
                db_file = self.get_db_file(conn, "production")
            finally:
                if conn:
                    conn.close()

            old_file_size = self.common_service.get_file_size(db_file)
            conn = None
            try:
                logger.info('Doing VACUUM on production database')
                if db_file:
                    conn = sqlite3.connect(db_file)
                    conn.execute("VACUUM;")
            except Exception as _:
                logger.exception('Error doing vacuum on production database')
            finally:
                if conn:
                    conn.close()

            new_file_size = self.common_service.get_file_size(db_file)

            logger.info('Production database file sizes. OLD: {}mb; NEW: {}mb'.format(old_file_size, new_file_size))

            logger.info("Finished [_vacuum_production_database]")
        except Exception as _:
            logger.exception("Error doing vacuum on production database")

    @staticmethod
    def get_db_file(conn, db_name_to_find):
        db_file = None
        pragma_cursor = conn.select("PRAGMA database_list;")
        for pragma_row in pragma_cursor:
            db_name, db_file = map(pragma_row.get_entry, (1, 2,))
            if db_name == db_name_to_find:
                break
        return db_file

    def _delete_fiscal(
        self,
        orders      # type: List[int]
    ):
        # type: (...) -> None

        orders_batches = [orders[x:x + 100] for x in xrange(0, len(orders), 100)]
        logger.info("Starting [_delete_orders]")
        for orders_batch in orders_batches:
            self.fiscal_repository.delete_fiscals(orders_batch)
        logger.info("Finished [_delete_orders]")

    def _backup_fiscal(self):
        # type: () -> List[int]

        logger.info("Starting [_backup_fiscal]")
        orders_stored = []

        is_timestamp = True
        end_date = self.common_service.get_end_date(is_timestamp)
        orders = self.fiscal_repository.get_orders(end_date)
        if len(orders) < 1:
            return []

        orders_by_month = self._split_by_months(
            items=orders,
            timestamp=is_timestamp
        )

        try:
            for month in orders_by_month:
                self.fiscal_repository.fill_fiscal_backup_database(
                    suffix_date=month,
                    orders=orders_by_month[month]
                )

            orders_stored = [o[0] for o in orders]
            logger.info("Finished [_backup_orders]")

        except Exception as _:
            logger.exception("Error doing backup")

        return orders_stored

    @staticmethod
    def _hard_clean_fiscal_database(db_file):
        logger.info("Starting [_hard_clean_fiscal_database]")
        subprocess.call(
            [
                "sqlite3",
                db_file,
                ".mode insert",
                ".output dump_all.sql",
                ".dump",
                ".read dump_all.sql"
            ],
            shell=True
        )
        logger.info("Finished [_hard_clean_fiscal_database]")


def execute_db_transaction(conn, query, close_conn=True):
    try:
        conn.transaction_start()
        conn.query('''BEGIN TRANSACTION;''')
        conn.query(query)
    except Exception as ex:
        conn.query('''ROLLBACK TRANSACTION;''')
        raise ex
    else:
        conn.query('''COMMIT TRANSACTION;''')
    finally:
        if close_conn:
            conn.close()



import logging
import xml.sax

from cache_functions import UsersInfoCache
from xml.etree import cElementTree as eTree

from typing import List

from messageprocessorutil import UuidMessageProcessor
from mwrepository import ProductRepository
from production.manager import ProductionManager
from production.model import ProductionOrder, Item

from ._SystemOrderParser import SystemOrderParser

performance_event_logger = logging.getLogger("ProductionPerformanceEvent")


class OrderModifiedProcessor(UuidMessageProcessor):
    def __init__(self, production_manager, product_repository, user_cache_thread=None):
        # type: (ProductionManager, ProductRepository, UsersInfoCache) -> None
        super(OrderModifiedProcessor, self).__init__()
        self.production_manager = production_manager
        self.part_code_jit_lines = product_repository.get_all_part_code_jit_lines()
        self.user_cache_thread = user_cache_thread

    def parse_data(self, data):
        self._performance_log(data)
        handler = SystemOrderParser(self.part_code_jit_lines, user_cache_thread=self.user_cache_thread)
        xml.sax.parseString(data, handler)
        if handler.order.order_id == 0:
            return None

        self._order_items(handler.order.items)
        return handler.order

    @staticmethod
    def _performance_log(
        data,   # type: str
    ):
        # type (...) -> None

        try:
            data_xml = eTree.XML(data)
            order_id = (data_xml.find(".//Order") or {}).get("orderId", "")
            subject = data_xml.get("subject", "")
            event_type = data_xml.get("type", "")
            message = "Event - {:10} / {:20} / {}".format(order_id, subject, event_type)
            performance_event_logger.info(message)
        except Exception as _:
            performance_event_logger.exception("Error parsing event xml")

    def _order_items(self, items):
        # type: (List[Item]) -> None

        for item in items:
            item.items = sorted(item.items, key=lambda i: (i.sequence, i.part_code))
            for son in item.items:
                if len(son.items) > 0:
                    self._order_items(son.items)

    def call_business(self, order):
        # type: (ProductionOrder) -> None
        if order is not None:
            self.production_manager.order_modified(order)

    def format_response(self, message_bus, message, event, input_obj, result):
        # no response
        pass

    def format_exception(self, message_bus, message, event, input_obj, exception):
        # no response
        pass

    def format_parse_exception(self, message_bus, message, event, data, exception):
        # no response
        pass



from typing import Dict, Optional, List

from messagebus import MessageBus, Message, DefaultToken
from messageprocessor import MessageProcessorMessageHandler, MessageProcessor, MessageProcessorCallback, \
    DefaultMessageProcessorExecutorFactory
from simplehttprouter import Route

from ._HttpMessageProcessor import HttpMessageProcessor


class HttpMessageProcessorMessageHandler(MessageProcessorMessageHandler):
    def __init__(self, route_processor, subject_counter=None, callbacks=None):
        # type: (Dict[Route, MessageProcessor], Optional[Dict[str, int]], Optional[List[MessageProcessorCallback]]) -> None # noqa
        msg_dict = {
            DefaultToken.TK_HTTP_MSG: HttpMessageProcessor(route_processor)
        }
        executor_factory = None
        if callbacks is not None:
            executor_factory = DefaultMessageProcessorExecutorFactory(callbacks)
        super(HttpMessageProcessorMessageHandler, self).__init__(None,
                                                                 msg_dict,
                                                                 executor_factory,
                                                                 subject_counter)

    def handle_message(self, message_bus, message):
        # type: (MessageBus, Message) -> None
        if message.token == DefaultToken.TK_HTTP_MSG:
            message.data = message
        super(HttpMessageProcessorMessageHandler, self).handle_message(message_bus, message)



from modelcommon.model import SitefResponse
from sysdefs import eft
from typing import (
    Any,
    Union
)

from model import (
    CancelPaymentOrderData,
    ConfirmFinishPaymentData,
    EftData,
    PaymentResponseMessage,
    TestConnectionData
)
from paymentprocessor import PaymentProcessor
from ._SitefPaymentProcessor import SitefPaymentProcessor


class SitefDebitPaymentProcessor(PaymentProcessor):
    def __init__(self, message_bus, dialog_manager, config, logger):
        super(SitefDebitPaymentProcessor, self).__init__(message_bus, dialog_manager, config, logger)
        self.sitef_processor = SitefPaymentProcessor(message_bus, dialog_manager, config, logger)
        self.allowed_payments = config.find_values("AllowedPayments") or []
        self.additional_data = self._get_additional_data()

    def connection_status(
        self,
        test_connection_data    # type: TestConnectionData
    ):
        # type: (...) -> bool

        return self.sitef_processor.connection_status(test_connection_data=test_connection_data)

    def add_payment(self, start_payment_data):
        ret = self.sitef_processor.add_payment(
            start_payment_data=start_payment_data,
            additional_params=self.additional_data,
        )
        self.register_eft_activity(EftData(start_payment_data.pos_id, start_payment_data.session_id, eft.EFT_CREDITSALE,
                                           0 if isinstance(ret, PaymentResponseMessage) else 1,
                                           start_payment_data.amount, orderid=start_payment_data.order_id))
        return ret

    def cancel_payment_order(
        self,
        cancel_payment_order_data   # type: CancelPaymentOrderData
    ):
        # type: (...) -> Any[Union[SitefResponse, str]]

        return self.sitef_processor.cancel_payment_order(cancel_payment_order_data=cancel_payment_order_data)

    def cancel_payment(self):
        return self.sitef_processor.cancel_payment()

    def finish_payment(self, finish_payment_data):
        return self.sitef_processor.finish_payment(finish_payment_data)

    def confirm_finish_payment(
        self,
        confirm_finish_payment_data     # type: ConfirmFinishPaymentData
    ):
        # type: (...) -> None

        pass

    def _get_additional_data(self):
        additional_data_list = []
        if self.allowed_payments:
            additional_data_list.append("{{TransacoesHabilitadas={}}}".format(';'.join(self.allowed_payments)))

        return ';'.join(additional_data_list)

    def get_name(self):
        return self.sitef_processor.get_name()

    def stop(self):
        self.sitef_processor.stop()



class ItemGoals(object):
    def __init__(self, name, part_codes, quantity_goal, quantity_operator_goal, quantity_sold):
        self.name = name
        self.part_codes = part_codes
        self.quantity_goal = quantity_goal
        self.quantity_operator_goal = quantity_operator_goal
        self.quantity_sold = quantity_sold



from logging import Logger

from interactor.route.payment.cancel.dto import (
    CancelPaymentRequest,
    CancelPaymentResponse
)
from model.config import Config
from model.repository import PosRepositoryModel
from validation_util import (
    check_if_business_day_is_opened,
    check_if_operator_is_logged,
    check_if_order_is_totaled,
    validate_api_key
)


class CancelPaymentInteractor(object):
    def __init__(
            self,
            logger,          # type: Logger
            config,          # type: Config
            pos_repository   # type: PosRepositoryModel
     ):
        # type: (...) -> None

        self.logger = logger
        self.config = config
        self.pos_repository = pos_repository

    def execute(
            self,
            request   # type: CancelPaymentRequest
    ):
        # type: (...) -> CancelPaymentResponse

        pos_id = validate_api_key(
            request=request,
            secure_api_key_list=self.config.api_key_mapping
        )

        model = self.pos_repository.get_model(pos_id)

        check_if_business_day_is_opened(
            repository=self.pos_repository,
            model=model
        )
        check_if_operator_is_logged(
            repository=self.pos_repository,
            model=model
        )
        check_if_order_is_totaled(
            repository=self.pos_repository,
            model=model
        )

        response = self.pos_repository.do_clear_tender(
            pos_id=pos_id,
            model=model,
            tender_id=request.tender_id
        )

        return CancelPaymentResponse(response)



import json
import logging

from helper import remove_accents
from old_helper import BaseRepository
from msgbus import MBEasyContext
from persistence import Connection
from typing import Dict

logger = logging.getLogger("FiscalWrapper")


class ProductRepository(BaseRepository):
    def __init__(self, mb_context):
        # type: (MBEasyContext) -> None

        super(ProductRepository, self).__init__(mb_context)

    def get_tender_types(self):
        # type: () -> Dict

        def inner_func(conn):
            # type: (Connection) -> Dict
            res = {}
            try:
                query = """SELECT TenderId, TenderDescr, CustomAttr FROM TenderType """

                for row in conn.select(query):
                    custom_attr = row.get_entry("CustomAttr")
                    if custom_attr in ['', None] or "tBand" not in custom_attr:
                        continue

                    tender_id = row.get_entry("TenderId")
                    description = remove_accents(row.get_entry("TenderDescr"))
                    custom_attr = json.loads(custom_attr)

                    res[description.upper()] = {
                        "tenderId": tender_id,
                        "description": description,
                        "cnpj": str(custom_attr.get("CNPJ", None)),
                        "clientId": str(custom_attr.get("clientId", None)),
                        "tBand": str(custom_attr.get("tBand", None)),
                        "tPag": str(custom_attr.get("tPag", None)),
                        "cAdmC": str(custom_attr.get("cAdmC", None)),
                    }

                return res
            except Exception as _:
                logger.exception('Error to get tenders types')
                return res

        return self.execute_with_connection(inner_func)




# encoding: utf-8

from ._BenefitsHeaderDto import BenefitsHeaderDto
from ._BenefitsBodyDto import BenefitsBodyDto


class BenefitsGeneratorDto(object):
    def __init__(self, header, body):
        # type: (BenefitsHeaderDto, BenefitsBodyDto) -> ()
        self.header = header
        self.body = body



# encoding: utf-8
import json
from datetime import (
    datetime,
    timedelta
)

from msgbus import MBEasyContext
from old_helper import BaseRepository
from persistence import Connection
from typing import (
    List,
    Dict
)

from tefcancellationreport.model import (
    TefTender,
    CardBrand
)

CARD_BRANDS_CACHE = None


class OrderRepository(BaseRepository):

    def __init__(
        self,
        mb_context,     # type: MBEasyContext
        pos_list        # type: List
    ):
        # type: (...) -> None

        super(
            OrderRepository,
            self
        ).__init__(mb_context)

        self.pos_list = pos_list

    def get_tef_tenders_by_pos(
        self,
        initial_date,   # type: datetime
        end_date        # type: datetime
    ):
        # type: (...) -> List[TefTender]

        def inner_func(
            conn    # type: Connection
        ):
            # type: (...)->List

            query = self._OrdersQuery.format(initial_date, end_date)
            tenders = {}
            for row in conn.select(query):
                tender_detail = json.loads(row.get_entry("TenderDetail"))
                receipt_cancel = tender_detail.get("ReceiptCancel")
                if receipt_cancel is not None:
                    if receipt_cancel.get("NsuHost") is not None:
                        continue

                date = row.get_entry("Timestamp")[:10]
                tef_tender = TefTender(
                    row_database=row,
                    tender_detail=tender_detail
                )
                if date in tenders:
                    tenders[date].append(tef_tender)
                else:
                    tenders[date] = [tef_tender]

            return [tenders]

        return self.execute_parallel_in_all_databases_returning_flat_list(inner_func, self.pos_list)

    def get_card_brands(self):
        # type: () -> Dict[int, CardBrand]

        global CARD_BRANDS_CACHE

        def inner_func(
            conn    # type: Connection
        ):
            # type: (...) -> Dict[int, CardBrand]

            query = self._CardBrandsQuery
            card_brands = {}
            for row in conn.select(query):
                card_brands[int(row.get_entry(0))] = CardBrand(row_database=row)

            return card_brands

        if not CARD_BRANDS_CACHE:
            CARD_BRANDS_CACHE = self.execute_with_connection(
                inner_func,
                service_name="FiscalPersistence"
            )

        return CARD_BRANDS_CACHE

    @staticmethod
    def _format_date(
        date    # type: datetime
    ):
        # type: (...) -> str

        return date.strftime("%Y-%m-%dT%H:%M:%S.%f")

    _OrdersQuery = """
                    SELECT
                        TT.TenderDescr,
                        OT.Timestamp,
                        OT.TenderAmount,
                        OT.TenderDetail
                    FROM TenderElectronicType AS TET
                    JOIN TenderType AS TT
                        ON TT.ElectronicType=TET.TypeId
                    JOIN OrderTender AS OT
                        ON OT.TenderId=TT.TenderId
                    JOIN Orders AS O
                        ON O.OrderId=OT.OrderId
                    WHERE TET.TypeId
                        NOT IN (CAST(0 AS INTEGER), CAST(9 AS INTEGER))
                        AND OT.TenderAmount=CAST(0 AS FLOAT)
                        AND O.BusinessPeriod BETWEEN '{0}' AND '{1}'
                    ORDER BY OT.Timestamp DESC
                  """

    _CardBrandsQuery = """
                    SELECT Bandeira, Descricao
                    FROM fiscal.BandeiraCartao
                  """



# -*- coding: utf-8 -*-

import json
import logging

from xml.etree.ElementTree import Element

from application.interactor import NDiscountInteractor, OrderInteractor
from msgbus import TK_SYS_ACK
from mw_helper import show_message_dialog, show_check_list_box_dialog, show_filtered_list_box_dialog
from posot import OrderTaker
from sysactions import get_model, get_current_order, format_amount, show_messagebox, get_posot, translate_message, \
    StopAction
from xml.etree import cElementTree as eTree

from typing import Dict, Optional, List

logger = logging.getLogger("PosActions")


def display_coupons(pos_id, table_id=None):
    model = get_model(pos_id)
    pos_ot = get_posot(model)
    order = get_current_order(model)
    if order is None:
        return

    try:
        benefits = get_benefits_from_order(pos_ot, order)
        if benefits is None or len(benefits) == 0:
            show_messagebox(pos_id, "$ORDER_HAS_NO_APPLIED_DISCOUNT")
            return

        applied_coupons = _get_applied_coupons_from_benefits(model, order, benefits, table_id)
        selected_indexes = _select_coupon(pos_id, applied_coupons)
        if not len(selected_indexes):
            return

        coupon_ids = []
        for selected_index in selected_indexes:
            index = int(selected_index)
            if index == 0:
                continue

            coupon = [x for x in applied_coupons if x["index"] == index]
            if not coupon:
                continue

            disable = coupon[0]["name"].split("##")[2]
            if not eval(disable):
                coupon_ids.append(coupon[0]["id"])

        msg = NDiscountInteractor.remove_benefits(pos_id, coupon_ids)
        if msg.token != TK_SYS_ACK:
            show_message_dialog(pos_id, "$ERROR", msg.data)
            raise StopAction()

        if table_id is not None:
            from table_actions import process_discount_in_table, clean_table_payments, create_table_modified_event

            table_picture = clean_table_payments(pos_id, table_id, delete_order_tender=False)
            create_table_modified_event(pos_id, table_picture)

            process_discount_in_table(pos_id, order, table_id)

        return True

    except StopAction:
        raise
    except Exception as ex:
        show_message_dialog(pos_id, "$ERROR", ex.message)
        logger.exception("Error removing discount")
        return False


def get_benefits_from_order(pos_ot, order, custom_properties=None):
    if not order:
        return []

    if not custom_properties:
        custom_properties = eTree.XML(pos_ot.getOrderCustomProperties(orderid=order.get("orderId")))

    benefit_list_custom_property = custom_properties.find(".//OrderProperty[@key='BENEFIT_LIST']")
    if benefit_list_custom_property is None:
        return []

    benefit_list_json = json.loads(benefit_list_custom_property.get("value"))
    benefit_json = benefit_list_json.get("benefitList")
    benefit_list = json.loads(benefit_json)
    benefit_list = [b for b in benefit_list if b['benefit_applied']]

    return benefit_list


def _get_applied_coupons_from_benefits(model, order, benefits, table_id):
    discounts = get_lines_discount_amounts_by_ids(order, table_id, None)
    all_coupons = translate_message(model, "ALL_COUPONS")
    applied_coupons = []
    total_amount = 0
    benefit_index = 1

    for benefit in benefits:
        benefit_applied = benefit.get("benefit_applied", False)
        if not benefit_applied:
            continue

        benefit_detail = benefit["benefit"]
        amount = 0
        if "used_discount_ids" in benefit:
            for d in benefit["used_discount_ids"]:
                amount += discounts[str(d)] if str(d) in discounts.keys() else 0
        if "order_tenders_ids" in benefit:
            for t in benefit["order_tenders_ids"]:
                amount += _get_tender_discount(t, discounts)
        if benefit_detail:
            added_sale_lines = len(benefit.get("added_sale_lines", []))
            if added_sale_lines and amount == 0:
                continue

            benefit_key = benefit["benefit_key"]
            description = (benefit_detail.get("userDescription") or benefit_detail.get("id") or "").encode("UTF-8")
            benefit_added_in_progress = benefit["order_state"] == "IN_PROGRESS"
            disable = benefit_added_in_progress or added_sale_lines > 0
            name = "{}##{}##{}".format(description, format_amount(model, amount, False), disable)
            data = {"index": benefit_index, "id": benefit_key, "name": name}
            applied_coupons.append(data)

        benefit_index += 1
        total_amount += amount

    name = "{}##{}##False".format(all_coupons, format_amount(model, total_amount, False))
    data = {"index": 0, "id": None, "name": name}
    applied_coupons.append(data)
    applied_coupons.sort(key=lambda x: (x["index"]))
    return applied_coupons


def _select_coupon(pos_id, coupons):
    options = ["{0}".format(c["name"]) for c in coupons]
    return show_check_list_box_dialog(pos_id, options, "$SELECT_ONE_COUPON", buttons="$REMOVE|$CANCEL")


def get_lines_discount_amounts_by_ids(order, table_id, discounts):
    if discounts is None:
        discounts = {}
    sale_lines = order.findall(".//SaleLine")
    for sale_line in sale_lines:
        if float(sale_line.get("qty") or 0) == 0:
            continue

        discounts_applied = sale_line.get("discountsApplied")
        if discounts_applied:
            _add_discounts_applied(discounts_applied, discounts)

    if table_id:
        discounts = _add_tenders_amounts_by_ids_from_table(int(order.get("posId", 0)), table_id, discounts)
    else:
        discounts = _add_tenders_amounts_by_ids_from_order(order, discounts)
    return discounts


def _add_tenders_amounts_by_ids_from_table(pos_id, table_id, discounts):
    # type (int, str, Dict) -> Dict

    from tablemgrapi import get_pos_ts
    model = get_model(pos_id)
    pos_ts = get_pos_ts(model)
    table_picture = pos_ts.getTablePicture(table_id)
    table_picture_xml = eTree.XML(table_picture)

    tender_lines = table_picture_xml.findall(".//ServiceTenders/ServiceTender")
    for tender_line in tender_lines:
        tender_id = tender_line.get("orderTenderId", None)
        _add_tender_discount(tender_line, discounts, tender_id)
    return discounts


def _add_tenders_amounts_by_ids_from_order(order, discounts):
    # type: (Element, Dict) -> Dict

    tender_lines = order.findall(".//Tender")
    for tender_line in tender_lines:
        tender_id = tender_line.get("tenderId", None)
        _add_tender_discount(tender_line, discounts, tender_id)
    return discounts


def _add_discounts_applied(discounts_applied, discounts):
    for discount in discounts_applied.split(','):
        discount_id, amount = discount.split(':')
        if discount_id != '2':
            if discount_id in discounts.keys():
                discounts[discount_id] += float(amount)
            else:
                discounts[discount_id] = float(amount)


def _add_tender_discount(tender_line, discounts, tender_id):
    # type: (Element, Dict, str) -> None

    if tender_id is None:
        return

    tender = "tender{}".format(tender_id)
    if tender not in discounts:
        discounts[tender] = 0

    discounts[tender] = float(tender_line.get("tenderAmount", None) or 0)


def _get_tender_discount(tender_id, discounts):
    tender_id = "tender{}".format(tender_id)
    if tender_id in discounts.keys():
        return discounts[tender_id]
    return 0


def get_all_order_tenders_ids(
        pos_ot,         # type: OrderTaker
        model,          # type: Element
        order,          # type: Element
        table_id=None   # type: Optional[str]
):
    # type: (...) -> (List[int], List[int])

    benefits = get_benefits_from_order(pos_ot, order) or []
    benefit_tenders_ids = []
    for b in benefits:
        for t in b["order_tenders_ids"]:
            benefit_tenders_ids.append(int(t))

    if table_id:
        all_tenders_ids = _get_table_service_tenders_by_order_tender(model, table_id)
        not_benefit_tenders_ids = [all_tenders_ids[k] for k in all_tenders_ids.keys() if k not in benefit_tenders_ids]
    else:
        all_tenders_ids = []
        tender_lines = order.findall(".//Tender")
        for tender_line in tender_lines:
            all_tenders_ids.append(int(tender_line.get("tenderId", -1)))

        not_benefit_tenders_ids = [k for k in all_tenders_ids if k not in benefit_tenders_ids]

    return not_benefit_tenders_ids, benefit_tenders_ids


def _get_table_service_tenders_by_order_tender(model, table_id):
    from tablemgrapi import get_pos_ts

    tenders_ids = {}
    pos_ts = get_pos_ts(model)
    table_picture = pos_ts.getTablePicture(table_id)
    table_picture_xml = eTree.XML(table_picture)
    service_tenders = table_picture_xml.findall(".//ServiceTenders/ServiceTender")
    for service_tender in service_tenders:
        tenders_ids[int(service_tender.get("orderTenderId") or -1)] = int(service_tender.get("serviceTenderId") or -1)

    return tenders_ids


def display_coupons_from_report(pos_id, initial_date, end_date, operator_id):
    # type: (str, str, str, Optional[str]) -> List

    try:
        if initial_date is None or end_date is None:
            raise StopAction()

        benefits = OrderInteractor(pos_id).get_orders_benefit(initial_date, end_date, operator_id)
        if benefits is None or len(benefits) == 1:
            show_messagebox(pos_id, "$ORDER_HAS_NO_APPLIED_DISCOUNT")
            raise StopAction()

        applied_benefits = [benefit["name"] for benefit in benefits]
        sorted_applied_benefits = [applied_benefits[0]] + sorted(list(dict.fromkeys(applied_benefits[1:])))
        options = ["{0}##True".format(benefit) for benefit in sorted_applied_benefits]
        selected_index = show_filtered_list_box_dialog(pos_id, options, "$BENEFIT_REPORT", mask="NOFILTER")

        if selected_index is None:
            raise StopAction()

        chosen_benefit = [x for x in benefits if x["name"] == sorted_applied_benefits[selected_index]]
        return chosen_benefit
    except StopAction:
        raise
    except Exception as ex:
        show_message_dialog(pos_id, "$ERROR", ex.message)
        logger.exception("Error display coupons from report")
        raise StopAction()




from logging import Logger

from model.config import Config
from model.error_type import ExternalServicesErrorsType
from model.exception import ExternalServiceErrors
from model.repository import PosRepositoryModel
from interactor.route.menu.get_products.dto import GetProductsRequest
from validation_util import validate_api_key
from messagebus import MessageBusException


class GetProductsInteractor(object):
    def __init__(self, logger, config, pos_repository, cache_menu_products):
        # type: (Logger, Config, PosRepositoryModel, bytearray) -> None

        self.logger = logger
        self.config = config
        self.pos_repository = pos_repository
        self.cache_menu_products = cache_menu_products

    def execute(self, request):
        # type: (GetProductsRequest) -> bytearray

        validate_api_key(request, self.config.api_key_mapping)
        if not self.cache_menu_products:
            try:
                self.cache_menu_products = self.pos_repository.get_menu_data()
            except MessageBusException:
                raise ExternalServiceErrors(ExternalServicesErrorsType.ServiceUnavailable)

        if not self.cache_menu_products:
            raise ExternalServiceErrors(ExternalServicesErrorsType.MenuDataNotAvailable)

        return bytearray(self.cache_menu_products)



from _CashReportDto import CashReportDto
from _CashReportBodyDto import CashReportBodyDto
from _CashReportFiltersDto import CashReportFiltersDto
from _CashReportBodyLineDto import CashReportBodyLineDto
from _Order import Order
from _OrderTender import OrderTender
from _OrderInfo import OrderInfo

__all__ = ["CashReportDto",
           "CashReportBodyDto",
           "CashReportFiltersDto",
           "CashReportBodyLineDto",
           "Order",
           "OrderTender",
           "OrderInfo"
           ]



from production.model import Item
from typing import List, Optional

from ._OrderChangerProductionBox import OrderChangerProductionBox


class HideOptionBox(OrderChangerProductionBox):
    def __init__(self, name, sons, fix_default_qty, logger):
        super(HideOptionBox, self).__init__(name, sons, logger)
        self.fix_default_qty = fix_default_qty

    def change_order(self, order):
        returned_items = []
        for item in order.items:
            new_sons = self.remove_option(item, None)
            returned_items.extend(new_sons)
        order.items = returned_items
        return order

    def remove_option(
        self,
        item,   # type: Item
        father  # type: Optional[Item]
    ):
        # type: (...) -> Optional[List[Item]]

        if item.item_type == "OPTION":
            new_items = self._get_new_items(father, item)

            returned_items = []
            for son in new_items:
                new_sons = self.remove_option(son, item)
                returned_items.extend(new_sons)

            return returned_items

        filtered_sons = []
        for son in item.items:
            filtered_sons.extend(self.remove_option(son, item))

        item.items = filtered_sons
        return [item]

    def _get_new_items(
        self,
        father,     # type: Item
        item        # type: Item
    ):
        # type: (...) -> List[Item]

        new_items = []
        for son in item.items:
            son.handle_remove_father(item)

            if self.fix_default_qty and self._father_product_or_combo(father) and item.default_qty > 0:
                son.default_qty = son.qty

            if son.default_qty == 0 and son.qty == 0:
                continue

            new_items.append(son)
        return new_items

    @staticmethod
    def _father_product_or_combo(father):
        return father and (father.item_type == "COMBO" or father.item_type == "PRODUCT" or father.item_type == "CANADD")



from httpmessageprocessorutil.automapped import auto_map
from model.order_detail import AdditionalProperty
from typing import List


@auto_map({"system_api_key": {"name": "X-Api-Key", "location": "header", "type": str},
           "pos_id": {"name": "posId", "type": str},
           "to_add_properties": {"name": "toAddProperties", "type": List[AdditionalProperty]}})
class AddPropertiesRequest(object):
    def __init__(self, system_api_key, pos_id, to_add_properties):
        # type: (str, str, str) -> None

        self.system_api_key = system_api_key
        self.pos_id = pos_id
        self.to_add_properties = to_add_properties



# encoding: utf-8
from datetime import datetime

from utils import get_store_id


class TefCancellationHeaderDto(object):

    def __init__(
        self,
        title,          # type: str
        start_date,     # type: datetime
        end_date        # type: datetime
    ):
        # type: (...) -> None

        self.title = title
        self.store_name = get_store_id()
        self.start_date = start_date
        self.end_date = end_date



import base64

from datetime import datetime, timedelta
from time import sleep

from typing import Dict
from xml.etree import cElementTree as eTree


from actions.eft import cancel_eft_transactions
from old_helper import convert_from_utf_to_localtime

import sysactions
from custom_sysactions import user_authorized_to_execute_action
from utils import get_void_reason_id, void_order, void_fiscal_order
from .. import logger
from ..util.get_menu_user_logged import get_menu_manager_authenticated_user

VOID_ATTEMPTS = 3


@sysactions.action
@user_authorized_to_execute_action
def do_void_order(
    pos_id,
    void_reason=None,
    fiscal_void='false',
    order_id="",
    show_cancel_message=True,
    is_mobile=False
):
    # type: (...) -> bool

    try:

        model = sysactions.get_model(pos_id)
        originator_pos_id = pos_id
        is_mobile = str(is_mobile).lower() == 'true'
        canceled_by = None

        if fiscal_void.lower() == 'true':
            order_id, originator_pos_id = _void_fiscal_order(pos_id, model)
            canceled_by = _get_canceled_by(
                pos_id,
                model,
            )
        else:
            sysactions.check_current_order(pos_id, model)

        void_reason = get_void_reason_id(
            pos_id=pos_id,
            selected_reason=void_reason,
            operator=canceled_by
        )
        if not void_reason:
            return True

        _void_order(
            order_id,
            originator_pos_id,
            void_reason,
        )
        _cancel_eft_transactions(
            pos_id,
            model,
            order_id,
            is_mobile,
            canceled_by,
        )
        _print_tef_receipts(
            pos_id,
            is_mobile,
        )

        if show_cancel_message:
            sysactions.show_messagebox(pos_id, "$ORDER_VOIDED_WITH_SUCCESS")

        return True
    except sysactions.StopAction as _:
        return False
    except Exception as _:
        logger.exception("Error voiding order {}".format(str(order_id) if order_id else 'undefined'))
        if show_cancel_message:
            sysactions.show_messagebox(pos_id, "$ERROR_CANCEL_ORDER")

        return False


def _void_order(
    order_id,           # type: str
    originator_pos_id,  # type: str
    void_reason,        # type: Dict
):
    # type: (...) -> None

    for _ in range(VOID_ATTEMPTS):
        try:
            void_order(originator_pos_id, order_id=order_id, void_reason=void_reason)
            break
        except Exception as _:
            logger.exception("Error voiding order")
            sleep(1)


def _print_tef_receipts(
    pos_id,     # type: str
    is_mobile,  # type: bool
):
    # type: (...) -> None

    if not is_mobile:
        from posactions import get_printer_receipt
        printer_receipt = get_printer_receipt()
        printer_receipt.print_tef_receipts(pos_id, cancel_receipt=True)


def _cancel_eft_transactions(
    pos_id,         # type: str
    model,          # type: eTree.XML
    order_id,       # type: str
    is_mobile,      # type: bool
    canceled_by,    # type: str
):
    # type: (...) -> None

    from posactions import is_table_service
    is_table_service = is_table_service(model)

    cancel_eft_transactions(
        pos_id,
        model,
        order_id,
        tender=None,
        is_table_service=is_table_service,
        is_mobile=is_mobile,
        operator_id=canceled_by
    )


def _get_canceled_by(
    pos_id,     # type: str
    model,      # type: eTree.XML
):
    # type: (...) -> str

    authenticated_manager = get_menu_manager_authenticated_user(int(pos_id))
    if authenticated_manager is not None:
        canceled_by = authenticated_manager
    else:
        operator = model.find("Operator")
        canceled_by = operator.get("id", "UNDEFINED") if operator is not None else "UNDEFINED"

    return canceled_by


def _void_fiscal_order(
    pos_id,  # type: str
    model,  # type: eTree.XML
):
    # type: (...) -> (str, str)

    order_id = ""
    pos_ot = sysactions.get_posot(model)
    limit_datetime = datetime.now() - timedelta(minutes=30)
    date_format = "%Y-%m-%dT%H:%M:%S"
    orders = pos_ot.listOrders(state="PAID", since=limit_datetime.strftime(date_format))
    order_options = []
    baseurl = "/mwapp/services/PosController/POS{}/?token=TK_POS_EXECUTEACTION&format=2&isBase64=true&payload=" \
        .format(pos_id)
    order_descr = "{} - {} POS{:02} - {:05} - R${:.2f}"
    for order in orders:
        if 'custom_properties' in order and 'FISCALIZATION_DATE' in order['custom_properties']:
            fiscalization_prop = order['custom_properties']['FISCALIZATION_DATE']
            fiscalization_date = convert_from_utf_to_localtime(datetime.strptime(fiscalization_prop, date_format))
            if fiscalization_date.replace(tzinfo=None) > limit_datetime:
                descr = order_descr.format(
                    fiscalization_date.strftime('%H:%M'),
                    order['podType'],
                    int(order['originatorId'][3:]),
                    int(order['orderId']),
                    float(order['totalGross'])
                )
                payload = base64.b64encode("\0".join(("doOrderPicture", pos_id, order['orderId'])))
                url = baseurl + payload
                order_options.append((order['orderId'], descr, url))

    if len(order_options) > 0:
        selected = sysactions.show_order_preview(pos_id,
                                                 order_options,
                                                 "$SELECT_AN_ORDER_TO_CANCEL",
                                                 buttons="CANCEL|OK",
                                                 onthefly=True)
        if (selected is None) or (selected[0] == "0"):
            raise sysactions.StopAction()
        order_id = selected[1]

    if order_id:
        originator_pos_id = int(next(iter([x for x in orders if x['orderId'] == order_id]))['originatorId'][3:])
        void_fiscal_order(originator_pos_id, order_id)
    else:
        sysactions.show_messagebox(pos_id, "$NO_ORDER_FOUND_TO_CANCEL")
        raise sysactions.StopAction()

    return order_id, originator_pos_id



from logging import Logger
from xml.etree import cElementTree as eTree
from xml.etree.ElementTree import Element

from application.model.exceptions import ProductNotFound
from typing import (
    Dict,
    List,
    Optional
)

from sysactions import get_model, get_posot, get_pricelist

from application.model.benefit import Benefit
from application.model.order import Order
from application.model.conditions import Restrictions
from application.repository import NDiscountRepository
from application.service import OrderService


class AddBenefitToApplyUnitOfWork(object):
    def __init__(
            self,
            logger,                     # type: Logger
            pos_id,                     # type: int
            benefit,                    # type: Benefit
            benefit_applier,            # type: str
            tender_id,                  # type: str
            order_picture,              # type: Element
            order_service,              # type: OrderService
            repository,                 # type: NDiscountRepository
            product_groups,             # type: Dict
            table_id=None,              # type: Optional[str]
            user_auth_level=False       # type: Optional[bool]
    ):
        # type: (...) -> None

        self.logger = logger
        self.pos_id = pos_id
        self.benefit = benefit
        self.benefit_key = None
        self.benefit_id = benefit.id
        self.raw_benefit = benefit.raw
        self.benefit_applier = benefit_applier
        self.tender_id = tender_id
        self.order_service = order_service
        self.repository = repository
        self.product_groups = product_groups
        self.added_sale_lines = []
        self.promotion_tenders = []
        self.order_tenders_ids = []
        self.benefit_order_info = None
        self.user_auth_level = user_auth_level

        self.model = get_model(pos_id)
        self.order = self.order_service.get_order(
            self.pos_id,
            create_order=any([a.name == "itemsToAdd" for a in self.benefit.actions]),
            order_picture=order_picture,
            model=self.model,
            table_id=table_id
        )

        self.pos_ot = get_posot(self.model)
        self.price_list = get_pricelist(self.model)

    def update_order(self):
        self.order = self.order_service.get_order(
            self.pos_id,
            create_order=any([a.name == "itemsToAdd" for a in self.benefit.actions])
        )

    def add_sale_line(self, part_code, is_free, price, tender_id, burnt_items, restriction, on_total, action_id):
        # type: (int, bool, float, int, Dict, Restrictions, bool, int) -> Dict

        if int(part_code) not in self.repository.get_all_products():
            raise ProductNotFound()

        try:
            self.pos_ot.blkopnotify = True
            added_line_number = self._sale_item(part_code, 1, on_total)
            self._overwrite_item_price(added_line_number, price, is_free)
            self.update_order()
            items_to_burn = self.order.get_items_to_burn(self.order, self.benefit, burnt_items, self.product_groups)
            required_items = Order.get_required_items(self.order, self.benefit)
        finally:
            self.pos_ot.blkopnotify = False

        if on_total:
            self._set_item_custom_properties(added_line_number, part_code, items_to_burn, required_items, restriction)
        else:
            self._set_item_custom_properties(added_line_number, part_code, items_to_burn, required_items)

        added_sale_lines = [x for x in self.order.sale_lines if int(x.line_number) == int(added_line_number)]
        tender_amount = self._get_tender_amount(added_sale_lines, is_free, price)

        if tender_amount > 0:
            self.promotion_tenders.append({"action_id": action_id, "amount": tender_amount, "tender_id": tender_id})

        self._get_item_to_burn_from_sale_lines(self.benefit, added_sale_lines, items_to_burn, part_code)
        return items_to_burn

    @staticmethod
    def _get_item_to_burn_from_sale_lines(benefit, added_sale_lines, items_to_burn, part_code):
        if not benefit.conditions.burn_items:
            return

        for sale_line in added_sale_lines:
            if sale_line.item_id == str(part_code):
                key = "{}.{}.{}.{}".format(sale_line.line_number, sale_line.level, sale_line.context, sale_line.item_id)
                if key in items_to_burn:
                    items_to_burn[key] += sale_line.qty
                else:
                    items_to_burn[key] = sale_line.qty

    def _set_item_custom_properties(self, added_sale_line_number, part_code, items_to_burn, required_items,
                                    restriction=None):
        # type: (int, int, Dict, List, Restrictions) -> None

        properties = {"benefitKey": self.benefit_key}
        self._set_discount_parent_sale_line(items_to_burn, required_items, properties)
        self._set_item_seat_and_service_period(items_to_burn, required_items, properties)
        self._set_item_sale_type(properties, restriction)
        self.pos_ot.setOrderCustomProperties(properties, "", added_sale_line_number, "1", "0", part_code)

    def _set_discount_parent_sale_line(self, items_to_burn, required_items, properties):
        # type: (Dict, List, Dict) -> None

        items = None
        if items_to_burn:
            items = [x for x in self.order.sale_lines if x.key == next(iter(items_to_burn))]
        elif len(required_items) > 0:
            items = [x for x in self.order.sale_lines if x.key == next(iter(required_items))]

        if items:
            line_number = items[0].line_number
            properties.update({"relationLineNumber": line_number})

    def _set_item_seat_and_service_period(self, items_to_burn, required_items, properties):
        # type: (Dict, List, Dict) -> None

        items = None
        if items_to_burn:
            items = [x for x in self.order.sale_lines if x.key == next(iter(items_to_burn))]
        elif len(required_items) > 0:
            items = [x for x in self.order.sale_lines if x.key == next(iter(required_items))]

        if items:
            seat = items[0].seat
            if seat:
                properties.update({"seat": seat})

            service_period = items[0].service_period.str_json
            if service_period:
                properties.update({"servicePeriod": service_period})

    def _set_item_sale_type(self, properties, restriction=None):
        # type: (Dict, Restrictions) -> None

        if restriction and restriction.valid_sale_types and len(restriction.valid_sale_types) > 0:
            sale_type = restriction.valid_sale_types[0]
        else:
            sale_type = self.order.sale_type if self.order.sale_type != "EAT_IN" else None

        if sale_type and self.order.table_id:
            properties.update({"saleType": sale_type})

    @staticmethod
    def _get_tender_amount(added_sale_lines, is_free, price):
        # type: (List, bool, float) -> float

        tender_amount = 0
        for sale_line in added_sale_lines:
            if price > 0 and is_free:
                tender_amount += sale_line.item_price

        return tender_amount

    def _overwrite_item_price(
            self,
            added_sale_line_number,   # type: int
            price,                    # type: Optional[float]
            is_free                   # type: bool
    ):
        # type (...) -> None

        price = 0 if is_free else price
        if price is not None:
            self.pos_ot.priceOverwrite(self.pos_id, added_sale_line_number, str(price))
            self.logger.info("Price overwritten. New Price: {}".format(str(price)))

    def _sale_item(self, item_id, quantity, on_total):
        # type (int, int, bool) -> int

        after_total = "1" if on_total else "0"
        context_id = "1.%s" % item_id

        params = [self.pos_id, context_id, self.price_list, quantity]
        sale_response = self.pos_ot.doSale(*params, aftertotal=after_total, applydefopts=True)
        added_sale_line = eTree.XML(sale_response)
        added_sale_line_number = int(added_sale_line.get("lineNumber"))

        self.added_sale_lines.append(added_sale_line_number)
        self.logger.info("New sale line added. LineNumber: {}; ItemId: {}".format(added_sale_line_number, context_id))

        return added_sale_line_number

    def set_benefit_key(self, benefit_key):
        self.benefit_key = benefit_key
